"""
Numba optimized linear algebra functions for 4x4 matrices and 2x2 matrices.
"""

from __future__ import absolute_import, print_function, division
from dtmm.conf import NCDTYPE, NFDTYPE, NUMBA_TARGET,NUMBA_PARALLEL, NUMBA_CACHE, NUMBA_FASTMATH, CDTYPE, FDTYPE
from numba import njit, prange, guvectorize, boolean
import numpy as np

if not NUMBA_PARALLEL:
    prange = range
    

@njit([NFDTYPE(NFDTYPE[:]),NFDTYPE(NCDTYPE[:])], cache=NUMBA_CACHE, fastmath=NUMBA_FASTMATH)        
def _vecabs2(v):
    """Computes vec.(vec.conj)"""
    out = 0.
    for i in range(len(v)):
        out = out + v[i].real**2 + v[i].imag**2
    return out

@njit([NFDTYPE(NFDTYPE[:]),NCDTYPE(NCDTYPE[:])], cache=NUMBA_CACHE, fastmath=NUMBA_FASTMATH)        
def _vnorm2(v):
    """Computes vec.vec"""
    return v[0]*v[0] + v[1]*v[1] + v[2]*v[2]

@njit([NFDTYPE[:](NFDTYPE[:],NFDTYPE[:],NFDTYPE[:]),NCDTYPE[:](NCDTYPE[:],NCDTYPE[:],NCDTYPE[:]), ], cache=NUMBA_CACHE, fastmath=NUMBA_FASTMATH)        
def _cross(v1,v2,v3):
    """performs vector cross product"""
    v30 = v1[1] * v2[2] - v1[2] * v2[1]
    v31 = v1[2] * v2[0] - v1[0] * v2[2]
    v32 = v1[0] * v2[1] - v1[1] * v2[0]
    v3[0] = v30
    v3[1] = v31
    v3[2] = v32
    return v3

_eigvec0_decl = [
                 (NFDTYPE,NFDTYPE,NFDTYPE,NFDTYPE,NFDTYPE,NFDTYPE,NFDTYPE,NFDTYPE[:],NFDTYPE[:],NFDTYPE[:],NFDTYPE[:]),
                 (NCDTYPE,NCDTYPE,NCDTYPE,NCDTYPE,NCDTYPE,NCDTYPE,NCDTYPE,NCDTYPE[:],NCDTYPE[:],NCDTYPE[:],NCDTYPE[:])
                 ]

@njit(_eigvec0_decl, cache=NUMBA_CACHE, fastmath=NUMBA_FASTMATH)            
def _eigvec0(a00,a11,a22,a01,a02,a12,eval0,vec0, tmp1, tmp2, tmp3):
    """The first eigenvector"""
    
    row0 = vec0
    row0[0] = a00 - eval0
    row0[1] = a01
    row0[2] = a02
   
    row1 = tmp1
    row1[0] = a01 
    row1[1] = a11 - eval0
    row1[2] = a12
    
    row2 = tmp2
    row2[0] = a02 
    row2[1] = a12
    row2[2] = a22 - eval0
    
    r0xr1 = _cross(row0, row1, tmp3 ) #tmp[0] is empty
    r1xr2 = _cross(row1, row2, tmp1 ) #row1 is in tmp1, but it can be deleted
    r0xr2 = _cross(row0, row2, tmp2 )
    
    d0 = _vecabs2(r0xr1)
    d1 = _vecabs2(r0xr2)
    d2 = _vecabs2(r1xr2)
    
    dmax = d0
    imax = 0
    
    if d1 > dmax:
        dmax = d1
        imax = 1
    if d2 > dmax:
        imax = 2
        
    if imax == 0:
        norm = 1/(_vnorm2(r0xr1)**0.5)
        vec0[:] = r0xr1[:] 
    elif imax == 1:
        norm = 1/(_vnorm2(r0xr2)**0.5)
        vec0[:] = r0xr2[:] 
    else:
        norm = 1/(_vnorm2(r1xr2)**0.5)
        vec0[:] = r1xr2[:] 
        
    vec0 *= norm

@njit([(NFDTYPE[:],NFDTYPE[:],NFDTYPE[:]),(NCDTYPE[:],NCDTYPE[:],NCDTYPE[:])], cache=NUMBA_CACHE, fastmath=NUMBA_FASTMATH) 
def _orthogonal_basis(w,u,v):
    """build an orthogonal basis based on normalized input vector w"""
    if abs(w[0]) > abs(w[1]):
        u[0] = (-w[2])
        u[1] = 0
        u[2] = w[0]
        scale = 1 / (_vecabs2(u)**0.5)
        u *= scale
    else:
        u[0] = 0
        u[1] = w[2]
        u[2] = (-w[1])
        scale = 1 / (_vecabs2(u)**0.5)
        u *= scale   
    _cross(w,u,v)

    
_eigvec1_decl = [
                 (NFDTYPE,NFDTYPE,NFDTYPE,NFDTYPE,NFDTYPE,NFDTYPE,NFDTYPE[:],NFDTYPE,NFDTYPE[:],NFDTYPE[:],NFDTYPE[:]),
                 (NCDTYPE,NCDTYPE,NCDTYPE,NCDTYPE,NCDTYPE,NCDTYPE,NCDTYPE[:],NCDTYPE,NCDTYPE[:],NCDTYPE[:],NCDTYPE[:])
                 ]   

@njit(_eigvec1_decl, cache=NUMBA_CACHE, fastmath=NUMBA_FASTMATH)            
def _eigvec1(a00,a11,a22,a01,a02,a12,vec0,eval1, vec1, tmp1, tmp2):
    """The second eigenvector"""
    u = vec1
    v = tmp1
    AU = tmp2
    
    _orthogonal_basis(vec0, u, v)
    
    AU[0] = a00 * u[0] + a01 * u[1] + a02 * u[2]
    AU[1] = a01 * u[0] + a11 * u[1] + a12 * u[2]
    AU[2] = a02 * u[0] + a12 * u[1] + a22 * u[2]
    
    m00 = u[0] * AU[0] + u[1] * AU[1] + u[2] * AU[2] - eval1
    
    AV = tmp2
   
    AV[0] = a00 * v[0] + a01 * v[1] + a02 * v[2]
    AV[1] = a01 * v[0] + a11 * v[1] + a12 * v[2]
    AV[2] = a02 * v[0] + a12 * v[1] + a22 * v[2]    
    
    m01 = u[0] * AV[0] + u[1] * AV[1] + u[2] * AV[2]
    m11 = v[0] * AV[0] + v[1] * AV[1] + v[2] * AV[2] - eval1
    
    absm00 = abs(m00)
    absm01 = abs(m01)
    absm11 = abs(m11)
    
    if absm00 >= absm11:
        m = max(absm00,absm01)
        if m > 0:
            if absm00 >= absm01:
                m01 /= m00 
                m00 = 1/((1+m01**2)**0.5)
                m01 *= m00
            else:
                m00 /= m01
                m01 = 1/((1+m00**2)**0.5) 
                m00 *= m01
            for i in range(3):
                vec1[i] = m01* u[i] - m00 * v[i] #vec1 is u so we are doing vec1 = m01 * u
    else:
        m = max(absm11,absm01)
        if m > 0:
            if absm11 >= absm01:
                m01 /= m11
                m11= 1/((1+m01**2)**0.5)
                m01 *= m11
            else:
                m11 /= m01
                m01 = 1/((1+m11**2)**0.5) 
                m11 *= m01
                
            for i in range(3):
                vec1[i] = m11* u[i] - m01 * v[i]   
 
    
@njit([(NFDTYPE[:],NFDTYPE[:,:],NFDTYPE[:],NFDTYPE[:,:]),(NCDTYPE[:],NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:,:])], cache = NUMBA_CACHE, fastmath=NUMBA_FASTMATH)
def _sort_eigvec(eps,r, epsout, rout):
    """Eigen modes sorting based on eigenvalues. Finds extraordinary axis,
    performs cyclic permutation of axes to move the extraordinary axis to 3"""
    e0 = eps[0]#.real #real part of the refractive index
    e1 = eps[1]#.real
    e2 = eps[2]#.real
    
    m2 = np.abs(e1-e0)
    m1 = np.abs(e2-e0)
    m0 = np.abs(e2-e1)
    
    #assume extraordinary is 2    
    i,j,k = 0,1,2
    
    #if we need to move eigenavlue to axis 2... do a cyclical permutation, to preserve right-handed coordinate system
    if m2 > m1 or m2 > m0:
        if m1 > m0:
            #extraordinary is 0
            i,j,k = 1,2,0
        else:
            #extraordinary is 1
            i,j,k = 2,0,1
            
    #perform sorting

    eps0 = eps[i]
    eps1 = eps[j]
    eps2 = eps[k]
    
    r00 = r[i,0]
    r01 = r[i,1]
    r02 = r[i,2]
    
    r10 = r[j,0]
    r11 = r[j,1]
    r12 = r[j,2]
    
    r20 = r[k,0]
    r21 = r[k,1]
    r22 = r[k,2]
    
    rout[0] = (r00,r01,r02)
    rout[1] = (r10,r11,r12)
    rout[2] = (r20,r21,r22)
          
    epsout[0] = eps0
    epsout[1] = eps1
    epsout[2] = eps2
        
    
_EIG_DECL = [(NFDTYPE[:,:],boolean[:],NFDTYPE[:],NFDTYPE[:,:]), (NCDTYPE[:,:],boolean[:],NCDTYPE[:], NCDTYPE[:,:])]         

@guvectorize(_EIG_DECL, '(m,m),()->(m),(m,m)', target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)   
def _tensor_eig(tensor, is_real, eig,vec):
    """Computes eigenvalues of a tensor using analytical noniterative algorithm
    adapted from A Robust Eigensolver for 3 × 3 Symmetric Matrices by 
    David Eberly @ Geometric Tools.
    
    Input array is overwritten.
    """
    
    #if norm of the offdiagonal elements is smaller than this, treat the
    #tensor as a diagonal tensor. This improves handling of real diagonal epsilon 
    #data treated as complex data, where the complex part introduces unwanted
    # rotations of the eigenframe, in particular when working with floats.
    normtol = 1e-12
    
    m1 = max(abs(tensor[0,0]),abs(tensor[1,1]))
    m2 = max(abs(tensor[2,2]),abs(tensor[0,1]))
    m3 = max(abs(tensor[0,2]),abs(tensor[1,2]))
    
    scale = max(max(m1,m2),m3)
    
    if scale == 0.:
        #zero matrix.. set eigenvalues to zero
        eig[0] = 0.
        eig[1] = 0.
        eig[2] = 0.
        vec[0] = (1,0,0)
        vec[1] = (0,1,0)
        vec[2] = (0,0,1)
    else:
        #precondition the matrix to avoid floating-point overflow
        scalem = 1 / scale
        a00 = tensor[0,0]*scalem
        a11 = tensor[1,1]*scalem
        a22 = tensor[2,2]*scalem
        a01 = tensor[0,1]*scalem
        a02 = tensor[0,2]*scalem
        a12 = tensor[1,2]*scalem
        
        q = (a00 + a11 + a22)/3
        
        b00 = a00 - q
        b11 = a11 - q
        b22 = a22 - q
            
        norm = a01**2 + a02**2 + a12**2
        
        if abs(norm) > normtol:
            p = ((b00**2 + b11**2 +b22**2 + 2 * norm)/6)**0.5
        
            c00 = b11 * b22 - a12 * a12
            c01 = a01 * b22 - a12 * a02
            c02 = a01 * a12 - b11 * a02
            
            half_det = (b00 * c00 - a01 * c01 + a02 * c02) / (p * p * p * 2)
            
            if is_real[0] == True:
                #for real data, half_det is between -1 and 1, to avoid possible rounding error, clamp it, just to make sure
                half_det = min(max(half_det.real, -1),1)
                
            phi = np.arccos(half_det)/3.
            
            #this number is closest to 2*np.pi/3 in float or double
            twoThirdsPi = NFDTYPE(2.09439510239319549)
            
            eig0 = np.cos(phi + twoThirdsPi) * 2
            eig2 = np.cos(phi) * 2
            eig1 = -(eig0 + eig2)
            
            #eigenvalues sorted by increasing 
            eig0 = (eig0 * p + q) 
            eig1 = (eig1 * p + q) 
            eig2 = (eig2 * p + q) 
            
            #now compute the eigenvectors
            if half_det.real >= 0:
                _eigvec0(a00,a11,a22,a01,a02,a12,eig2,vec[2], vec[1],vec[0], eig)    
                _eigvec1(a00,a11,a22,a01,a02,a12,vec[2],eig1,vec[1],vec[0],eig)
                _cross(vec[1],vec[2],vec[0])
            else:
                _eigvec0(a00,a11,a22,a01,a02,a12,eig0,vec[0],vec[1],vec[2],eig) 
                _eigvec1(a00,a11,a22,a01,a02,a12,vec[0],eig1,vec[1],vec[2],eig) 
                _cross(vec[0],vec[1],vec[2])
                
            eig[0] = (eig0 *  scale) 
            eig[1] = (eig1 *  scale) 
            eig[2] = (eig2 *  scale) 
        else:
            #matrix is diagonal
            eig[0] = a00 *  scale
            eig[1] = a11 *  scale
            eig[2] = a22 *  scale
            vec[0] = (1,0,0)
            vec[1] = (0,1,0)
            vec[2] = (0,0,1)
        
        _sort_eigvec(eig,vec,eig,vec)

_EIG_DECL = [(NFDTYPE[:,:],NFDTYPE[:],NFDTYPE[:,:]), (NCDTYPE[:,:],NCDTYPE[:], NCDTYPE[:,:])]         


@guvectorize(_EIG_DECL, '(m,m)->(m),(m,m)', target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)   
def _eig(tensor, eig,vec):
    e,v = np.linalg.eig(tensor)
    #_sort_eigvec uses rows for eigvecs.. so we need to transpose v and vec
    v = v.transpose()
    _sort_eigvec(e,v,eig,vec.transpose())
    
def eig(matrix,overwrite_x = False):
    """Computes eigenvalues and eigenvectors of 3x3 matrix using numpy.linalg.eig.
    Eigenvalues are sorted so that eig[2] is the most distinct (extraordinary).

    Parameters
    ----------
    matrix: (...,3,3) array 
        A 3x3 matrix.
    overwrite_x : bool, optional
        Ifset, the function will write eigenvectors as rows in the input array.
        
    Returns
    -------
    w : (..., 3) array
        The eigenvalues, each repeated according to its multiplicity.
        The eigenvalues are ordered so that the third eigenvalue is most
        distinct and first two are least distinct
    
    v : (..., 3, 3) array
        The normalized (unit "length") eigenvectors, such that the
        column ``v[:,i]`` is the eigenvector corresponding to the
        eigenvalue ``w[i]``.
    """
    matrix = np.asarray(matrix)
    if matrix.shape[-2:] != (3,3):
        raise ValueError("Only 3x3 matrices supported")
    if overwrite_x == True:
        out = (np.empty(matrix.shape[:-1],matrix.dtype), matrix)
        return _eig(matrix,out[0], out[1])
    else:
        return _eig(matrix)   
        
def tensor_eig(tensor, overwrite_x = False):
    """Computes eigenvalues and eigenvectors of a tensor.
    
    Eigenvalues are sorted so that eig[2] is the most distinct (extraordinary).
    
    If tensor is provided as a length 6 matrix, the elements are 
    a[0,0], a[1,1], a[2,2], a[0,1], a[0,2], a[1,2]. If provided as a (3x3) 
    matrix a, the rest of the elements are silently ignored.
    
    Parameters
    ----------
    tensor : (...,6) or (...,3,3) array 
        A length 6 array or 3x3 matrix
    overwrite_x : bool, optional
        If tensor is (...,3,3) array, the function will write eigenvectors
        as rows in the input array.
        
    Returns
    -------
    w : (..., 3) array
        The eigenvalues, each repeated according to its multiplicity.
        The eigenvalues are ordered so that the third eigenvalue is most
        distinct and first two are least distinct
    
    v : (..., 3, 3) array
        The normalized (unit "length") eigenvectors, such that the
        column ``v[:,i]`` is the eigenvector corresponding to the
        eigenvalue ``w[i]``.
    """
    tensor = np.asarray(tensor)
    matrix = tensor
    out = None
    
    _is_real = (np.iscomplexobj(matrix) == False)
    
    eig = _tensor_eig
    
    if tensor.shape[-1] == 6:
        matrix = np.empty(tensor.shape[0:-1] + (3,3),tensor.dtype)
        matrix[...,0,0] = tensor[...,0]
        matrix[...,1,1] = tensor[...,1]
        matrix[...,2,2] = tensor[...,2]
        matrix[...,0,1] = tensor[...,3]
        matrix[...,0,2] = tensor[...,4]
        matrix[...,1,2] = tensor[...,5]
        # we can overwrite temporary matrix data
        out = (np.empty(tensor.shape[:-1] + (3,),tensor.dtype), matrix)
    elif matrix.shape[-2:] != (3,3):
        raise ValueError("Not a matrix of size (...,3,3) or (...,6)")
    else:
        if overwrite_x == True:
            out = (np.empty(matrix.shape[:-1],matrix.dtype), matrix)
    if out is None:
        eig, vec = eig(matrix,_is_real)
    else:
        eig, vec = eig(matrix, _is_real, out[0], out[1])
        
    return eig, np.swapaxes(vec, -1,-2) #as returned by np.linalg.eig, eigenvectors are in columns, not rows
    

@njit([(NCDTYPE[:, :], NCDTYPE[:, :])], cache=NUMBA_CACHE, fastmath=NUMBA_FASTMATH)
def _inv2x2(src, dst):
    """

    Parameters
    ----------
    src : array
        (2, 2) input array to invert
    dst : array
        (2, 2) output array to store the invert input array


    """
    # Extract individual elements
    a = src[0, 0]
    b = src[0, 1]
    c = src[1, 0]
    d = src[1, 1]

    # Calculate determinant
    det = a * d - b * c
    if det == 0.:
        det = 0.
    else:
        det = 1. / det

    # Calcualte inverse
    dst[0, 0] =  d * det
    dst[0, 1] = -b * det
    dst[1, 0] = -c * det
    dst[1, 1] =  a * det

@njit([NCDTYPE[:,:](NCDTYPE[:,:],NCDTYPE[:,:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _inv4x4(src,dst):
    
    #calculate pairs for first 8 elements (cofactors)
    tmp0 = src[2,2] * src[3,3]
    tmp1 = src[3,2] * src[2,3]
    tmp2 = src[1,2] * src[3,3]
    tmp3 = src[3,2] * src[1,3]
    tmp4 = src[1,2] * src[2,3]
    tmp5 = src[2,2] * src[1,3]
    tmp6 = src[0,2] * src[3,3]
    tmp7 = src[3,2] * src[0,3]
    tmp8 = src[0,2] * src[2,3]
    tmp9 = src[2,2] * src[0,3]
    tmp10 = src[0,2] * src[1,3]
    tmp11 = src[1,2] * src[0,3]
    # calculate first 8 elements (cofactors)
    dst[0,0] = tmp0*src[1,1] + tmp3*src[2,1] + tmp4*src[3,1] -tmp1*src[1,1] - tmp2*src[2,1] - tmp5*src[3,1]
    dst[0,1] = tmp1*src[0,1] + tmp6*src[2,1] + tmp9*src[3,1] - tmp0*src[0,1] - tmp7*src[2,1] - tmp8*src[3,1]
    dst[0,2] = tmp2*src[0,1] + tmp7*src[1,1] + tmp10*src[3,1] - tmp3*src[0,1] - tmp6*src[1,1] - tmp11*src[3,1]
    dst[0,3] = tmp5*src[0,1] + tmp8*src[1,1] + tmp11*src[2,1] - tmp4*src[0,1] - tmp9*src[1,1] - tmp10*src[2,1]
    
    dst[1,0] = tmp1*src[1,0] + tmp2*src[2,0] + tmp5*src[3,0] - tmp0*src[1,0] - tmp3*src[2,0] - tmp4*src[3,0]
    dst[1,1] = tmp0*src[0,0] + tmp7*src[2,0] + tmp8*src[3,0] - tmp1*src[0,0] - tmp6*src[2,0] - tmp9*src[3,0]
    dst[1,2] = tmp3*src[0,0] + tmp6*src[1,0] + tmp11*src[3,0] - tmp2*src[0,0] - tmp7*src[1,0] - tmp10*src[3,0]
    dst[1,3] = tmp4*src[0,0] + tmp9*src[1,0] + tmp10*src[2,0] - tmp5*src[0,0] - tmp8*src[1,0] - tmp11*src[2,0]
    # calculate pairs for second 8 elements (cofactors) 
    tmp0 = src[2,0]*src[3,1]
    tmp1 = src[3,0]*src[2,1]
    tmp2 = src[1,0]*src[3,1]
    tmp3 = src[3,0]*src[1,1]
    tmp4 = src[1,0]*src[2,1]
    tmp5 = src[2,0]*src[1,1]
    tmp6 = src[0,0]*src[3,1]
    tmp7 = src[3,0]*src[0,1]
    tmp8 = src[0,0]*src[2,1]
    tmp9 = src[2,0]*src[0,1]
    tmp10 = src[0,0]*src[1,1]
    tmp11 = src[1,0]*src[0,1]

    dst[2,0] = tmp0*src[1,3] + tmp3*src[2,3] + tmp4*src[3,3] - (tmp1*src[1,3] + tmp2*src[2,3] + tmp5*src[3,3])
    dst[2,1] = tmp1*src[0,3] + tmp6*src[2,3] + tmp9*src[3,3] - (tmp0*src[0,3] + tmp7*src[2,3] + tmp8*src[3,3])
    dst[2,2] = tmp2*src[0,3] + tmp7*src[1,3] + tmp10*src[3,3] - (tmp3*src[0,3] + tmp6*src[1,3] + tmp11*src[3,3])
    dst[2,3] = tmp5*src[0,3] + tmp8*src[1,3] + tmp11*src[2,3] - (tmp4*src[0,3] + tmp9*src[1,3] + tmp10*src[2,3])
    dst[3,0] = tmp2*src[2,2] + tmp5*src[3,2] + tmp1*src[1,2] - (tmp4*src[3,2] + tmp0*src[1,2] + tmp3*src[2,2])
    dst[3,1] = tmp8*src[3,2] + tmp0*src[0,2] + tmp7*src[2,2] - (tmp6*src[2,2] + tmp9*src[3,2] + tmp1*src[0,2])
    dst[3,2] = tmp6*src[1,2] + tmp11*src[3,2] + tmp3*src[0,2] - (tmp10*src[3,2] + tmp2*src[0,2] + tmp7*src[1,2])
    dst[3,3] = tmp10*src[2,2] + tmp4*src[0,2] + tmp9*src[1,2] - (tmp8*src[1,2] + tmp11*src[2,2] + tmp5*src[0,2])
    #/* calculate determinant */
    det=src[0,0]*dst[0,0]+src[1,0]*dst[0,1]+src[2,0]*dst[0,2]+src[3,0]*dst[0,3]
    #/* calculate matrix inverse */
    if det == 0:
        det = 0.
    else:
        det = 1./det
    for i in range(4):
        for j in range(4):
            dst[i,j] = dst[i,j]*det
            
    return dst


@guvectorize([(NCDTYPE[:,:], NCDTYPE[:,:])], '(n,n)->(n,n)', target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def inv(mat, out):
    """inv(mat), gufunc
    
    Calculates inverse of a 4x4 complex matrix or 2x2 complex matrix
    
    Parameters
    ----------
    mat : ndarray
       Input array
    
    Examples
    --------
    
    >>> a = np.random.randn(4,4) + 0j
    >>> ai = inv(a)
    
    >>> from numpy.linalg import inv
    >>> ai2 = inv(a)
    
    >>> np.allclose(ai2,ai)
    True
    """
    if mat.shape[0] == 2:
        _inv2x2(mat,out)
    elif mat.shape[0] == 4:
        _inv4x4(mat,out)
    else:
        inv = np.linalg.inv(mat)
        out[...] = inv


@njit([(NCDTYPE[:,:],NCDTYPE[:,:],NCDTYPE[:,:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)    
def _dotmm4(a,b,out):
    a0 = a[0,0]*b[0,0]+a[0,1]*b[1,0]+a[0,2]*b[2,0]+a[0,3]*b[3,0]
    a1 = a[0,0]*b[0,1]+a[0,1]*b[1,1]+a[0,2]*b[2,1]+a[0,3]*b[3,1]
    a2 = a[0,0]*b[0,2]+a[0,1]*b[1,2]+a[0,2]*b[2,2]+a[0,3]*b[3,2]
    a3 = a[0,0]*b[0,3]+a[0,1]*b[1,3]+a[0,2]*b[2,3]+a[0,3]*b[3,3] 

    b0 = a[1,0]*b[0,0]+a[1,1]*b[1,0]+a[1,2]*b[2,0]+a[1,3]*b[3,0]
    b1 = a[1,0]*b[0,1]+a[1,1]*b[1,1]+a[1,2]*b[2,1]+a[1,3]*b[3,1]
    b2 = a[1,0]*b[0,2]+a[1,1]*b[1,2]+a[1,2]*b[2,2]+a[1,3]*b[3,2]
    b3 = a[1,0]*b[0,3]+a[1,1]*b[1,3]+a[1,2]*b[2,3]+a[1,3]*b[3,3] 

    c0 = a[2,0]*b[0,0]+a[2,1]*b[1,0]+a[2,2]*b[2,0]+a[2,3]*b[3,0]
    c1 = a[2,0]*b[0,1]+a[2,1]*b[1,1]+a[2,2]*b[2,1]+a[2,3]*b[3,1]
    c2 = a[2,0]*b[0,2]+a[2,1]*b[1,2]+a[2,2]*b[2,2]+a[2,3]*b[3,2]
    c3 = a[2,0]*b[0,3]+a[2,1]*b[1,3]+a[2,2]*b[2,3]+a[2,3]*b[3,3]
    
    d0 = a[3,0]*b[0,0]+a[3,1]*b[1,0]+a[3,2]*b[2,0]+a[3,3]*b[3,0]
    d1 = a[3,0]*b[0,1]+a[3,1]*b[1,1]+a[3,2]*b[2,1]+a[3,3]*b[3,1]
    d2 = a[3,0]*b[0,2]+a[3,1]*b[1,2]+a[3,2]*b[2,2]+a[3,3]*b[3,2]
    d3 = a[3,0]*b[0,3]+a[3,1]*b[1,3]+a[3,2]*b[2,3]+a[3,3]*b[3,3]     
 
    out[0,0] = a0
    out[0,1] = a1
    out[0,2] = a2
    out[0,3] = a3

    out[1,0] = b0
    out[1,1] = b1
    out[1,2] = b2
    out[1,3] = b3  

    out[2,0] = c0
    out[2,1] = c1
    out[2,2] = c2
    out[2,3] = c3 
    
    out[3,0] = d0
    out[3,1] = d1
    out[3,2] = d2
    out[3,3] = d3 
    
@njit([(NCDTYPE[:,:],NCDTYPE[:,:],NCDTYPE[:,:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)    
def _dotmm2(a,b,out):
    a0 = a[0,0]*b[0,0]+a[0,1]*b[1,0]
    a1 = a[0,0]*b[0,1]+a[0,1]*b[1,1]

    b0 = a[1,0]*b[0,0]+a[1,1]*b[1,0]
    b1 = a[1,0]*b[0,1]+a[1,1]*b[1,1] 
   
    out[0,0] = a0
    out[0,1] = a1

    out[1,0] = b0
    out[1,1] = b1
 
@njit([(NCDTYPE[:,:],NCDTYPE[:,:],NCDTYPE[:,:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)    
def _dotmm(a,b,out):
    tmp = np.zeros_like(out)
    for i in range(a.shape[0]):
        for j in range(b.shape[1]):
            #tmp = NCDTYPE(0)
            for k in range(a.shape[1]):
                tmp[i,j] += a[i,k] *b[k,j]
                
    out[...] = tmp


@njit([(NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)    
def _dotmv(a,b,out):
    tmp = np.zeros_like(out)
    for i in range(a.shape[0]):
        for k in range(a.shape[1]):
            tmp[i] += a[i,k] *b[k]
                
    out[...] = tmp

#@njit([(NCDTYPE[:,:],NFDTYPE[:],NCDTYPE[:,:])], cache = NUMBA_CACHE)    
#def _dotmr2(a,b,out):
#    a0 = a[0,0]*b[0]-a[0,2]*b[1]
#    a1 = a[0,1]*b[0]+a[0,3]*b[1]
#    a2 = a[0,0]*b[1]+a[0,2]*b[0]
#    a3 = -a[0,1]*b[1]+a[0,3]*b[0] 
#
#    b0 = a[1,0]*b[0]-a[1,2]*b[1]
#    b1 = a[1,1]*b[0]+a[1,3]*b[1]
#    b2 = a[1,0]*b[1]+a[1,2]*b[0]
#    b3 = -a[1,1]*b[1]+a[1,3]*b[0] 
#
#    c0 = a[2,0]*b[0]-a[2,2]*b[1]
#    c1 = a[2,1]*b[0]+a[2,3]*b[1]
#    c2 = a[2,0]*b[1]+a[2,2]*b[0]
#    c3 = -a[2,1]*b[1]+a[2,3]*b[0] 
#    
#    d0 = a[3,0]*b[0]-a[3,2]*b[1]
#    d1 = a[3,1]*b[0]+a[3,3]*b[1]
#    d2 = a[3,0]*b[1]+a[3,2]*b[0]
#    d3 = -a[3,1]*b[1]+a[3,3]*b[0]     
# 
#    out[0,0] = a0
#    out[0,1] = a1
#    out[0,2] = a2
#    out[0,3] = a3
#
#    out[1,0] = b0
#    out[1,1] = b1
#    out[1,2] = b2
#    out[1,3] = b3  
#
#    out[2,0] = c0
#    out[2,1] = c1
#    out[2,2] = c2
#    out[2,3] = c3 
#    
#    out[3,0] = d0
#    out[3,1] = d1
#    out[3,2] = d2
#    out[3,3] = d3 
#    
    
@njit([(NFDTYPE[:],NCDTYPE[:,:],NCDTYPE[:,:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)    
def _dotr2m(r,a,out):
    a0 = a[0,0]*r[0]-a[2,0]*r[1]
    a1 = a[0,1]*r[0]-a[2,1]*r[1]
    a2 = a[0,2]*r[0]-a[2,2]*r[1]
    a3 = a[0,3]*r[0]-a[2,3]*r[1] 

    b0 = a[1,0]*r[0]+a[3,0]*r[1]
    b1 = a[1,1]*r[0]+a[3,1]*r[1]
    b2 = a[1,2]*r[0]+a[3,2]*r[1]
    b3 = a[1,3]*r[0]+a[3,3]*r[1] 

    c0 = a[0,0]*r[1]+a[2,0]*r[0]
    c1 = a[0,1]*r[1]+a[2,1]*r[0]
    c2 = a[0,2]*r[1]+a[2,2]*r[0]
    c3 = a[0,3]*r[1]+a[2,3]*r[0] 
    
    d0 = -a[1,0]*r[1]+a[3,0]*r[0]
    d1 = -a[1,1]*r[1]+a[3,1]*r[0]
    d2 = -a[1,2]*r[1]+a[3,2]*r[0]
    d3 = -a[1,3]*r[1]+a[3,3]*r[0]   
 
    out[0,0] = a0
    out[0,1] = a1
    out[0,2] = a2
    out[0,3] = a3

    out[1,0] = b0
    out[1,1] = b1
    out[1,2] = b2
    out[1,3] = b3  

    out[2,0] = c0
    out[2,1] = c1
    out[2,2] = c2
    out[2,3] = c3 
    
    out[3,0] = d0
    out[3,1] = d1
    out[3,2] = d2
    out[3,3] = d3 
    
    
@njit([(NFDTYPE[:],NCDTYPE[:],NCDTYPE[:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)    
def _dotr2v(r,a,out):
    out[0]= a[0]*r[0]-a[2]*r[1]
    out[1]= a[1]*r[0]+a[3]*r[1]
    out[2]= a[0]*r[1]+a[2]*r[0]    
    out[3]= -a[1]*r[1]+a[3]*r[0]

    
@njit([(NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _dotmv4(a, b, out):
    out0 = a[0,0] * b[0] + a[0,1] * b[1] + a[0,2] * b[2] +a[0,3] * b[3]
    out1 = a[1,0] * b[0] + a[1,1] * b[1] + a[1,2] * b[2] +a[1,3] * b[3]
    out2 = a[2,0] * b[0] + a[2,1] * b[1] + a[2,2] * b[2] +a[2,3] * b[3]
    out3 = a[3,0] * b[0] + a[3,1] * b[1] + a[3,2] * b[2] +a[3,3] * b[3]
    out[0]= out0
    out[1]= out1
    out[2]= out2
    out[3]= out3
    
@njit([(NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _dotmv2(a, b, out):
    out0 = a[0,0] * b[0] + a[0,1] * b[1] 
    out1 = a[1,0] * b[0] + a[1,1] * b[1]
    out[0]= out0
    out[1]= out1
    

    
@njit([(NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:,:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _dotmd4(a, b, out):
    a0 = a[0,0]*b[0]
    a1 = a[0,1]*b[1]
    a2 = a[0,2]*b[2]
    a3 = a[0,3]*b[3] 
    
    out[0,0] = a0
    out[0,1] = a1
    out[0,2] = a2
    out[0,3] = a3

    a0 = a[1,0]*b[0]
    a1 = a[1,1]*b[1]
    a2 = a[1,2]*b[2]
    a3 = a[1,3]*b[3] 
    
    out[1,0] = a0
    out[1,1] = a1
    out[1,2] = a2
    out[1,3] = a3  

    a0 = a[2,0]*b[0]
    a1 = a[2,1]*b[1]
    a2 = a[2,2]*b[2]
    a3 = a[2,3]*b[3] 
    
    out[2,0] = a0
    out[2,1] = a1
    out[2,2] = a2
    out[2,3] = a3    
 
    a0 = a[3,0]*b[0]
    a1 = a[3,1]*b[1]
    a2 = a[3,2]*b[2]
    a3 = a[3,3]*b[3] 
    
    out[3,0] = a0
    out[3,1] = a1
    out[3,2] = a2
    out[3,3] = a3
    
@njit([(NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:,:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _dotmd2(a, b, out):
    a0 = a[0,0]*b[0]
    a1 = a[0,1]*b[1]
    
    out[0,0] = a0
    out[0,1] = a1

    a0 = a[1,0]*b[0]
    a1 = a[1,1]*b[1]
  
    out[1,0] = a0
    out[1,1] = a1

#@njit([(NCDTYPE[:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:])],parallel = NUMBA_PARALLEL, cache = NUMBA_CACHE)
#def _dotm1f4(a, b, out):
#    for i in prange(b.shape[1]):
#        for j in range(b.shape[2]):
#            b0 = b[0,i,j]
#            b1 = b[1,i,j]
#            b2 = b[2,i,j]
#            b3 = b[3,i,j]
#            
#            out0 = a[0,0] * b0 + a[0,1] * b1 + a[0,2] * b2 +a[0,3] * b3
#            out1 = a[1,0] * b0 + a[1,1] * b1 + a[1,2] * b2 +a[1,3] * b3
#            out2 = a[2,0] * b0 + a[2,1] * b1 + a[2,2] * b2 +a[2,3] * b3
#            out3 = a[3,0] * b0 + a[3,1] * b1 + a[3,2] * b2 +a[3,3] * b3
#            
#            out[0,i,j]= out0
#            out[1,i,j]= out1
#            out[2,i,j]= out2
#            out[3,i,j]= out3
#            
#@njit([(NCDTYPE[:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:])],parallel = NUMBA_PARALLEL, cache = NUMBA_CACHE)
#def _dotm1f2(a, b, out):
#    for i in prange(b.shape[1]):
#        for j in range(b.shape[2]):
#            b0 = b[0,i,j]
#            b1 = b[1,i,j]
#            
#            out0 = a[0,0] * b0 + a[0,1] * b1
#            out1 = a[1,0] * b0 + a[1,1] * b1 
#            
#            out[0,i,j]= out0
#            out[1,i,j]= out1
#            
@njit([(NCDTYPE[:,:,:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:])],parallel = NUMBA_PARALLEL, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _dotmf4(a, b, out):
    for i in prange(b.shape[1]):
        for j in range(b.shape[2]):
            b0 = b[0,i,j]
            b1 = b[1,i,j]
            b2 = b[2,i,j]
            b3 = b[3,i,j]
            
            out0 = a[i,j,0,0] * b0 + a[i,j,0,1] * b1 + a[i,j,0,2] * b2 +a[i,j,0,3] * b3
            out1 = a[i,j,1,0] * b0 + a[i,j,1,1] * b1 + a[i,j,1,2] * b2 +a[i,j,1,3] * b3
            out2 = a[i,j,2,0] * b0 + a[i,j,2,1] * b1 + a[i,j,2,2] * b2 +a[i,j,2,3] * b3
            out3 = a[i,j,3,0] * b0 + a[i,j,3,1] * b1 + a[i,j,3,2] * b2 +a[i,j,3,3] * b3
            
            out[0,i,j]= out0
            out[1,i,j]= out1
            out[2,i,j]= out2
            out[3,i,j]= out3  
            
@njit([(NCDTYPE[:,:,:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:])],parallel = NUMBA_PARALLEL, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _dotmf2(a, b, out):
    for i in prange(b.shape[1]):
        for j in range(b.shape[2]):
            b0 = b[0,i,j]
            b1 = b[1,i,j]
            
            out0 = a[i,j,0,0] * b0 + a[i,j,0,1] * b1 
            out1 = a[i,j,1,0] * b0 + a[i,j,1,1] * b1 
            
            out[0,i,j]= out0
            out[1,i,j]= out1


#@njit([(NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:])],cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
#def _dotmdmv4(a, d, b, f,out):
#    f0 = f[0]
#    f1 = f[1]
#    f2 = f[2]
#    f3 = f[3,]
#    
#    out0 = b[0,0] * f0 + b[0,1] * f1 + b[0,2] * f2 +b[0,3] * f3
#    out1 = b[1,0] * f0 + b[1,1] * f1 + b[1,2] * f2 +b[1,3] * f3
#    out2 = b[2,0] * f0 + b[2,1] * f1 + b[2,2] * f2 +b[2,3] * f3
#    out3 = b[3,0] * f0 + b[3,1] * f1 + b[3,2] * f2 +b[3,3] * f3
#    
#    b0 = out0*d[0]
#    b1 = out1*d[1]
#    b2 = out2*d[2]
#    b3 = out3*d[3]
#    
#    out[0]= a[0,0] * b0 + a[0,1] * b1 + a[0,2] * b2 +a[0,3] * b3
#    out[1]= a[1,0] * b0 + a[1,1] * b1 + a[1,2] * b2 +a[1,3] * b3
#    out[2]= a[2,0] * b0 + a[2,1] * b1 + a[2,2] * b2 +a[2,3] * b3
#    out[3]= a[3,0] * b0 + a[3,1] * b1 + a[3,2] * b2 +a[3,3] * b3   
#
#
#@njit([(NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
#def _dotmdmv2(a, d, b, f,out):
#    f0 = f[0]
#    f1 = f[1]
#    
#    out0 = b[0,0] * f0 + b[0,1] * f1
#    out1 = b[1,0] * f0 + b[1,1] * f1
#
#    
#    b0 = out0*d[0]
#    b1 = out1*d[1]
#    
#    out[0]= a[0,0] * b0 + a[0,1] * b1
#    out[1]= a[1,0] * b0 + a[1,1] * b1 
 
            
@njit([(NCDTYPE[:,:,:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:])],parallel = NUMBA_PARALLEL, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _dotmdmf4(a, d, b, f,out):
    for i in prange(f.shape[1]):
        for j in range(f.shape[2]):
            f0 = f[0,i,j]
            f1 = f[1,i,j]
            f2 = f[2,i,j]
            f3 = f[3,i,j]
            
            out0 = b[i,j,0,0] * f0 + b[i,j,0,1] * f1 + b[i,j,0,2] * f2 +b[i,j,0,3] * f3
            out1 = b[i,j,1,0] * f0 + b[i,j,1,1] * f1 + b[i,j,1,2] * f2 +b[i,j,1,3] * f3
            out2 = b[i,j,2,0] * f0 + b[i,j,2,1] * f1 + b[i,j,2,2] * f2 +b[i,j,2,3] * f3
            out3 = b[i,j,3,0] * f0 + b[i,j,3,1] * f1 + b[i,j,3,2] * f2 +b[i,j,3,3] * f3
            
            b0 = out0*d[i,j,0]
            b1 = out1*d[i,j,1]
            b2 = out2*d[i,j,2]
            b3 = out3*d[i,j,3]
            
            out[0,i,j]= a[i,j,0,0] * b0 + a[i,j,0,1] * b1 + a[i,j,0,2] * b2 +a[i,j,0,3] * b3
            out[1,i,j]= a[i,j,1,0] * b0 + a[i,j,1,1] * b1 + a[i,j,1,2] * b2 +a[i,j,1,3] * b3
            out[2,i,j]= a[i,j,2,0] * b0 + a[i,j,2,1] * b1 + a[i,j,2,2] * b2 +a[i,j,2,3] * b3
            out[3,i,j]= a[i,j,3,0] * b0 + a[i,j,3,1] * b1 + a[i,j,3,2] * b2 +a[i,j,3,3] * b3   

@njit([(NCDTYPE[:,:,:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:])],parallel = NUMBA_PARALLEL, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _dotmdmf2(a, d, b, f,out):
    for i in prange(f.shape[1]):
        for j in range(f.shape[2]):
            f0 = f[0,i,j]
            f1 = f[1,i,j]
            
            out0 = b[i,j,0,0] * f0 + b[i,j,0,1] * f1 
            out1 = b[i,j,1,0] * f0 + b[i,j,1,1] * f1 
            
            b0 = out0*d[i,j,0]
            b1 = out1*d[i,j,1]
            
            out[0,i,j]= a[i,j,0,0] * b0 + a[i,j,0,1] * b1 
            out[1,i,j]= a[i,j,1,0] * b0 + a[i,j,1,1] * b1  
                        

#@guvectorize([(NCDTYPE[:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:])],"(n,n),(n,m,k)->(n,m,k)",target = "cpu", cache = NUMBA_CACHE)
#def dotm1f(a, b, out):
#    if b.shape[0] == 2:
#        _dotm1f2(a, b, out)
#    else:
#        assert a.shape[0] >= 4 #make sure it is not smaller than 4
#        _dotm1f4(a, b, out)

@guvectorize([(NCDTYPE[:,:,:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:])],"(m,k,n,n),(n,m,k)->(n,m,k)",target = "cpu", cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _dotmf(a, b, out):
    """dotmf(a, b)
    
Computes a dot product of an array of 4x4 (or 2x2) matrix with 
a field array or an E-array (in case of 2x2 matrices).
"""
    if b.shape[0] == 2:
        _dotmf2(a, b, out)
    else:
        assert b.shape[0] >= 4 #make sure it is not smaller than 4
        _dotmf4(a, b, out)


def broadcast_m(m, field):
    """Broadcasts matrix m to match field spatial indices (the last two axes) so
    that it can be used in dot functions."""
    shape = m.shape[:-4]+ field.shape[-2:] + m.shape[-2:]
    return np.broadcast_to(m, shape)

def broadcast_d(d, field):
    """Broadcasts diagonal matrix d to match field spatial indices (the last two axes) so
    that it can be used in dot functions."""
    shape = d.shape[:-3]+ field.shape[-2:] + d.shape[-1:]
    return np.broadcast_to(d, shape)

def dotmf(a,b, out = None):
    """dotmf(a, b)
    
Computes a dot product of an array of 4x4 (or 2x2) matrix with 
a field array or an E-array (in case of 2x2 matrices).
"""
    a = np.asarray(a)
    b = np.asarray(b)
    a = broadcast_m(a, b)
    return _dotmf(a, b, out)

       
@guvectorize([(NCDTYPE[:,:,:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:,:],NCDTYPE[:,:,:],NCDTYPE[:,:,:])],"(m,k,n,n),(m,k,n),(m,k,n,n),(n,m,k)->(n,m,k)",target = "cpu", cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _dotmdmf(a, d,b,f, out):
    if f.shape[0] == 2:
        _dotmdmf2(a, d,b,f, out)
    else:    
        assert f.shape[0] >= 4 #make sure it is not smaller than 4
        _dotmdmf4(a,d, b, f,out)
        
def dotmdmf(a,d,b,f, out = None):
    """dotmdmf(a, d, b, f)
    
Computes a dot product of an array of 4x4 (or 2x2) matrices, array of diagonal matrices, 
another array of matrices and a field array or an E-array (in case of 2x2 matrices).

Notes
-----
This is equivalent to

>>> dotmf(dotmdm(a,d,b),f)
"""
    try:
        return _dotmdmf(a, d,b,f, out)
    except:
        m = dotmdm(a,d,b)
        return dotmf(m,f, out)
        
#    a = broadcast_m(a, f)
#    d = broadcast_d(d, f)
#    b = broadcast_m(b, f)
#    return _dotmdmf(a, d,b,f, out)
    

#@guvectorize([(NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:])],"(n,n),(n),(n,n),(n)->(n)",target =NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
#def dotmdmv(a, d,b,f, out):
#    """dotmdmv(a, d, b, f)
#    
#Computes a dot product of an array of 4x4 (or 2x2) matrices, array of diagonal matrices, 
#another array of matrices and a vector.
#
#Notes
#-----
#This is equivalent to
#
#>>> dotmv(dotmdm(a,d,b),f)
#"""
#    if f.shape[0] == 2:
#        _dotmdmv2(a, d,b,f, out)
#    else:    
#        assert f.shape[0] >= 4 #make sure it is not smaller than 4
#        _dotmdmv4(a,d, b, f,out)
                
                
@guvectorize([(NCDTYPE[:,:],NCDTYPE[:,:],NCDTYPE[:,:])],"(n,k),(k,m)->(n,m)",target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def dotmm(a, b, out):
    """dotmm(a, b)
    
Computes an efficient dot product of a 4x4,  2x2 
or a less efficient general matrix multiplication.
"""
    if a.shape[0] == 2 and  a.shape[1] == 2:
        _dotmm2(a, b, out)
    elif a.shape[0] == 4 and a.shape[1] == 4:
        _dotmm4(a, b, out)
    else:
        _dotmm(a, b, out)
#
#@guvectorize([(NCDTYPE[:,:,:,:],NCDTYPE[:,:,:,:],NCDTYPE[:,:,:,:])],"(n,n,k,k),(n,n,k,k)->(n,n,k,k)", cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)        
#def bdotmm3(m1,m2, out):
#    """Performs a dot product of two nxn block matrices of blocks 4x4.
#    Matrices must be of shape nxnx4x4 that describe two mxm matrices
#    (m = 4*n) of blocks of size 4x4
#    """
#    assert m1.shape[3] == 4
#    
#    tmp = np.empty(out.shape[2:],out.dtype)
#
#    n = m1.shape[0]    
#    for i in prange(n):
#        for j in range(n):
#            for k in range(n):
#                if k == 0:
#                    _dotmm4(m1[i,k], m2[k,j], out[i,j] )
#                else:
#                    _dotmm4(m1[i,k], m2[k,j], tmp)
#                    out[i,j] += tmp
#
#@njit([(NCDTYPE[:,:,:,:],NCDTYPE[:,:,:,:],NCDTYPE[:,:,:,:])], parallel = NUMBA_PARALLEL)        
#def bdotmm2(m1,m2, out):
#    """Performs a dot product of two nxn block matrices of blocks 4x4.
#    Matrices must be of shape nxnx4x4 that describe two mxm matrices
#    (m = 4*n) of blocks of size 4x4
#    """
#    assert m1.shape[3] == 4
#    
#    tmp = np.empty(out.shape[2:],out.dtype)
#
#    n = m1.shape[0]    
#    for i in prange(n):
#        for j in range(n):
#            for k in range(n):
#                if k == 0:
#                    _dotmm4(m1[i,k], m2[k,j], out[i,j] )
#                else:
#                    _dotmm4(m1[i,k], m2[k,j], tmp)
#                    out[i,j] += tmp


def bdotmm(m1,m2, out = None):
    """Performs a dot product of two nxn block matrices of blocks of size 4x4.
    Matrices must be of shape nxnx4x4 that describe two mxm matrices
    (m = 4*n) of blocks of size 4x4. 
    """
    assert m1.shape == m2.shape
    assert m2 is not out
    if out is None:
        out = np.empty(shape = m1.shape, dtype = CDTYPE)
        
    tmp = np.empty(shape = m1.shape, dtype = CDTYPE)

    for j in range(m1.shape[0]):
        dotmm(m1[j][:,None,:,:],m2, out = tmp)
        out[j] = tmp.sum(-4, out = out[j])
    return out

def _bdotmm_ref(m1,m2, out = None):
    """same as bdotmm, but slower, for testing"""
    m1,m2 = np.broadcast_arrays(m1,m2)
    if out is None:
        out = np.zeros(m1.shape, CDTYPE)
    else:
        out[...] = 0.
    n = m1.shape[-3] 
    assert n == m1.shape[-4]
    for i in range(n):
        for j in range(n):
            for k in range(n):
                out[...,i,j,:,:] += dotmm(m1[...,i,k,:,:],m2[...,k,j,:,:])
    return out


def bdotmd(m1,m2, out = None):
    """Performs a dot product of two nxn block matrices of blocks 4x4.
    The second matrix is block diagonal matrix of shape nx4x4,
    The first matrix must be of shape nxnx4x4 that describe two mxm matrices
    (m = 4*n) of blocks of size 4x4.
    """
    if out is None:
        out = np.empty(shape = m1.shape, dtype = CDTYPE)

    for j in range(m1.shape[-3]):
        dotmm(m1[j],m2,out[j])
    return out

def _bdotmd_ref(m1,d):
    """same as bdotmd, but slower, for testing"""
    out = np.zeros_like(m1)
    for i in range(m1.shape[0]):
        for k in range(m1.shape[0]):
            out[i,k] +=  np.dot(m1[i,k],d[k])
    return out


def bdotdm(m1,m2, out = None):
    """Performs a dot product of two nxn block matrices of blocks 4x4.
    The second matrix is block diagonal matrix of shape nx4x4,
    The first matrix must be of shape nxnx4x4 that describe two mxm matrices
    (m = 4*n) of blocks of size 4x4.
    """
    assert m2 is not out
    if out is None:
        out = np.empty(shape = m2.shape, dtype = CDTYPE)

    for j in range(m2.shape[-3]):
        dotmm(m1,m2[:,j],out[:,j])
    return out

def _bdotdm_ref(d,m):
    """Same as bdotdm, but slower, for testing"""
    out = np.zeros_like(m)
    for j in range(m.shape[0]):
        for k in range(m.shape[0]):
            out[k,j] +=  np.dot(d[k],m[k,j])
    return out


def bdotmv(m,v, out = None):
    """Performs a dot product of two nxn block matrices of blocks 4x4.
    The second matrix is block diagonal matrix of shape nx4x4,
    The first matrix must be of shape nxnx4x4 that describe two mxm matrices
    (m = 4*n) of blocks of size 4x4.
    """
    tmp = dotmv(m,v)
    return tmp.sum(-2, out = out)


def _bdotmv_ref(m,v):
    """Performs a dot product of two nxn block matrices of blocks 4x4.
    Matrices must be of shape nxnx4x4 that describe two mxm matrices
    (m = 4*n) of blocks of size 4x4
    """
    
    m = np.moveaxis(m,-2,-3).copy()
    shape = v.shape
    m = m.reshape(m.shape[0]*4,m.shape[0]*4).copy()
    v = v.reshape(v.shape[0]*4).copy()
    out = dotmv(m,v).copy()
    out = out.reshape(shape).copy()
    return out



        
@guvectorize([(NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:,:])],"(n,n),(n)->(n,n)",target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def dotmd(a, d, out):
     """dotmd(a, d)
     
Computes a dot product of a 4x4 (or 2x2) matrix with a diagonal
matrix represented by a vector of shape 4 (or 2).
"""
     if a.shape[0]== 2:
         _dotmd2(a, d, out)
     else:
         assert a.shape[0] >= 4 #make sure it is not smaller than 4
         _dotmd4(a, d, out)

@guvectorize([(NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:])],"(n,n),(n)->(n)",target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def dotmv(a, b, out):
    """dotmv(a, b)
    
Computes a dot product of a 4x4 or 2x2 matrix with a vector.
"""
    if a.shape[0] == 2:
        _dotmv2(a, b, out)
    elif a.shape[0] == 4:
        assert a.shape[0] >= 4 #make sure it is not smaller than 4
        _dotmv4(a, b, out)
    else:
        _dotmv(a, b, out)
    
@guvectorize([(NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:,:],NCDTYPE[:,:])],"(n,n),(n),(n,n)->(n,n)",target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def dotmdm(a, d, b, out):
    """dotmdm(a, d, b)
    
Computes a dot product of a 4x4 (or 2x2) matrix with a diagonal matrix (4- or 2-vector) 
and another 4x4 (or 2x2) matrix.
"""
    #assert a.shape[0] >= 4 #make sure it is not smaller than 4 
    if a.shape[0] == 2:
        _dotmd2(a, d, out)
        _dotmm2(out,b,out)
    else:
        assert a.shape[0] >= 4 #make sure it is not smaller than 4
        _dotmd4(a, d, out)
        _dotmm4(out,b,out)        
    

def multi_dot(arrays,  axis = 0, reverse = False):
    """Computes dot product of multiple 2x2 or 4x4 matrices. If reverse is 
    specified, it is performed in reversed order. Axis defines the axis over 
    which matrices are multiplied."""
    out = None
    if axis != 0:
        arrays = np.asarray(arrays)
        indices = range(arrays.shape[axis])
        arrays = np.rollaxis(arrays, axis)
        #arrays = np.moveaxis(arrays, axis)
    else:
        indices = range(len(arrays))
    if reverse == True:
        indices = reversed(indices)
    for i in indices:
        if out is None:
            out = np.asarray(arrays[i]).copy()
        else:
            b = np.broadcast(out, arrays[i])
            if b.shape == out.shape:
                out = dotmm(out, arrays[i], out = out)
            else:
                out = dotmm(out, arrays[i])
    return out
    
    
__all__ = ["inv", "dotmm","dotmf","dotmv","dotmdm","dotmd","multi_dot","eig","tensor_eig"]




"""
Director and optical data creation and IO functions.

Conversion functions
--------------------

* :func:`.director2data` generates optical data from director.
* :func:`.Q2data` generates optical data from the Q tensor.
* :func:`.director2order` computes order parameter from directo length.
* :func:`.director2Q` computes uniaxial Q tensor.
* :func:`.Q2director` computes an effective uniaxial director from a biaxial Q tensor.
* :func:`.director2angles` computes Euler rotation angles from the director.
* :func:`.angles2director` computes the director from Euler rotation angles.
* :func:`.Q2eps` converts Q tensor to epsilon tensor.
* :func:`.eps2epsva` converts epsilon tensor to epsilon eigenvalues and Euler angles.
* :func:`.epsva2eps` converts epsilon eigenvalues and Euler angles to epsilon tensor.
* :func:`.tensor2matrix` convert tensor to matrix.
* :func:`.matrix2tensor` convert matrix to tensor.
* :func:`.refind2eps` convert refractive index eigenvalues to epsilon eigenvalues.
* :func:`.uniaxial_order` creates uniaxial tensor from a biaxial eigenvalues.
* :func:`.eig_symmetry` creates effective tensor of given symmetry.
* :func:`.effective_data` computes effective (mean layers) 1D data from 3D data.

IO functions
------------

* :func:`.read_director` reads 3D director data.
* :func:`.read_tensor` reads 3D tensor data.
* :func:`.read_raw` reads any raw data.
* :func:`.load_stack` loads optical data.
* :func:`.save_stack` saves optical data.

Data creation
-------------

* :func:`.nematic_droplet_data` builds sample data.
* :func:`.nematic_droplet_director` builds sample director.
* :func:`.cholesteric_droplet_data` builds sample data.
* :func:`.cholesteric_director` builds sample director.
* :func:`.expand` expands director data.
* :func:`.sphere_mask` builds a masking array.
* :func:`.rot90_director` rotates director by 90 degrees
* :func:`.rotate_director` rotates data by any angle
"""

from __future__ import absolute_import, print_function, division

import numpy as np
import numba
import sys

from dtmm.conf import FDTYPE, CDTYPE, NFDTYPE, NCDTYPE, NUMBA_CACHE,\
NF32DTYPE,NF64DTYPE,NC128DTYPE,NC64DTYPE, DTMMConfig
from dtmm.rotation import rotation_matrix_x,rotation_matrix_y,rotation_matrix_z, rotate_vector, rotation_angles, rotation_matrix, rotate_diagonal_tensor
from dtmm.wave import betaphi, k0
from dtmm.fft import fft2, ifft2
from dtmm.linalg import tensor_eig

def read_director(file, shape, dtype = FDTYPE,  sep = "", endian = sys.byteorder, order = "zyxn", nvec = "xyz"):
    """Reads raw director data from a binary or text file. 
    
    A convinient way to read director data from file. 
    
    Parameters
    ----------
    file : str or file
        Open file object or filename.
    shape : sequence of ints
        Shape of the data array, e.g., ``(50, 24, 34, 3)``
    dtype : data-type
        Data type of the raw data. It is used to determine the size of the items 
        in the file.
    sep : str
        Separator between items if file is a text file.
        Empty ("") separator means the file should be treated as binary.
        Spaces (" ") in the separator match zero or more whitespace characters.
        A separator consisting only of spaces must match at least one
        whitespace.
    endian : str, optional
        Endianess of the data in file, e.g. 'little' or 'big'. If endian is 
        specified and it is different than sys.endian, data is byteswapped. 
        By default no byteswapping is done. 
    order : str, optional
        Data order. It can be any permutation of 'xyzn'. Defaults to 'zyxn'. It
        describes what are the meaning of axes in data.
    nvec : str, optional
        Order of the director data coordinates. Any permutation of 'x', 'y' and 
        'z', e.g. 'yxz', 'zxy' ... 
    """
    try:
        i,j,k,c = shape
    except:
        raise TypeError("shape must be director data shape (z,y,x,n)")
    data = read_raw(file, shape, dtype, sep = sep, endian = endian)
    return raw2director(data, order, nvec)

def read_tensor(file, shape, dtype = FDTYPE,  sep = "", endian = sys.byteorder, order = "zyxn"):
    """Reads raw tensor data from a binary or text file. 
    
    A convinient way to read tensor data from file.

    
    Parameters
    ----------
    file : str or file
        Open file object or filename.
    shape : sequence of ints
        Shape of the data array, e.g., ``(50, 24, 34, 6)``
    dtype : data-type
        Data type of the raw data. It is used to determine the size of the items 
        in the file.
    sep : str
        Separator between items if file is a text file.
        Empty ("") separator means the file should be treated as binary.
        Spaces (" ") in the separator match zero or more whitespace characters.
        A separator consisting only of spaces must match at least one
        whitespace.
    endian : str, optional
        Endianess of the data in file, e.g. 'little' or 'big'. If endian is 
        specified and it is different than sys.endian, data is byteswapped. 
        By default no byteswapping is done. 
    order : str, optional
        Data order. It can be any permutation of 'xyzn'. Defaults to 'zyxn'. It
        describes what are the meaning of axes in data.
    """
    try:
        i,j,k,c = shape
    except:
        raise TypeError("shape must be 3D tensor data shape (z,y,x,n)")
    data = read_raw(file, shape, dtype, sep = sep, endian = endian)
    return raw2director(data, order) #no swapping of Q tensor elements, so we can use raw2director

def rotate_director(rmat, data, method = "linear",  fill_value = (0.,0.,0.), norm = True, out = None):
    """
    Rotate a director field around the center of the compute box by a specified
    rotation matrix. This rotation is lossy, as datapoints are interpolated.
    The shape of the output remains the same.
    
    Parameters
    ----------
    rmat : array_like
        A 3x3 rotation matrix.
    data: array_like
        Array specifying director field with ndim = 4
    method : str
        Interpolation method "linear" or "nearest"
    fill_value : numbers, optional
        If provided, the values (length 3 vector) to use for points outside of the
        interpolation domain. Defaults to (0.,0.,0.).
    norm : bool,
        Whether to normalize the length of the director to 1. after rotation 
        (interpolation) is performed. Because of interpolation error, the length 
        of the director changes slightly, and this options adds a constant 
        length constraint to reduce the error.
    out : ndarray, optional
        Output array.
        
    Returns
    -------
    y : ndarray
        A rotated director field
        
    See Also
    --------   
    data.rot90_director : a lossless rotation by 90 degrees.
        
    """
    
    from scipy.interpolate import RegularGridInterpolator
    verbose_level = DTMMConfig.verbose
    if verbose_level >0:
        print("Rotating director.")   
        
    out = np.empty_like(data)
    nz,ny,nx,nv = data.shape
    shape = (nz,ny,nx)
    az, ay, ax = [np.arange(-l / 2. + .5, l / 2. + .5) for l in shape]
    
    fillx, filly, fillz = fill_value
    xdir = RegularGridInterpolator((az, ay, ax), data[...,0], 
              fill_value = fillx,bounds_error = False, method = method)
    ydir = RegularGridInterpolator((az, ay, ax), data[...,1], 
              fill_value = filly,bounds_error = False, method = method)
    zdir = RegularGridInterpolator((az, ay, ax), data[...,2], 
              fill_value = fillz,bounds_error = False, method = method)
    zz,yy,xx = np.meshgrid(az,ay,ax, indexing = "ij", copy = False, sparse = True)

    out[...,0] = xx
    out[...,1] = yy
    out[...,2] = zz

    out = rotate_vector(rmat.T,out, out) #rotate coordinates

    #out2 = out.copy()
    #out2[...,0] = out[...,2]
    #out2[...,2] = out[...,0]
    
    out2 = out[...,::-1] #reverse direction instead of copying
    
    #interpolate new director field
    xnew = xdir(out2) 
    ynew = ydir(out2)
    znew = zdir(out2)
    
    out[...,0] = xnew 
    out[...,1] = ynew
    out[...,2] = znew
    
    out = rotate_vector(rmat,out, out) #rotate vector in each voxel
    
    if norm == True:
        s = director2order(out)
        mask = (s == 0.)
        s[mask] = 1.
        return np.divide(out, s[...,None], out)
    return out

def rot90_director(data,axis = "+x", out = None):
    """
    Rotate a director field by 90 degrees around the specified axis.
    
    Parameters
    ----------
    data: array_like
        Array specifying director field with ndim = 4.
    axis: str
        Axis around which to perform rotation. Can be in the form of
        '[s][n]X' where the optional parameter 's' can be "+" or "-" decribing 
        the sign of rotation. [n] is an integer describing number of rotations 
        to perform, and 'X' is one of 'x', 'y' 'z', and defines rotation axis.
    out : ndarray, optional
        Output array.
    
    Returns
    -------
    y : ndarray
        A rotated director field
        
    See Also
    --------   
    data.rotate_director : a general rotation for arbitrary angle.        
    """
    nz,ny,nx,nv = data.shape

    axis_name = axis[-1]
    try:
        k = int(axis[:-1])
    except ValueError:
        k = int(axis[:-1]+"1")
    angle = np.pi/2*k
    if axis_name == "x":
        r = rotation_matrix_x(angle)
        axes = (1,0)
    elif axis_name == "y":
        r = rotation_matrix_y(angle)
        axes = (0,2)
    elif axis_name == "z":
        r = rotation_matrix_z(angle)
        axes = (2,1)
    else:
        raise ValueError("Unknown axis type {}".format( axis_name))
    data_rot = np.rot90(data,k = k, axes = axes)#first rotate data points
    return rotate_vector(r,data_rot,out)#rotate vector in each voxel
    
    
def director2data(director, mask = None, no = 1.5, ne = 1.6, nhost = None,scale_factor = 1.,
                  thickness = None):
    """Builds optical data from director data. Director length is treated as
    an order parameter. Order parameter of S=1 means that refractive indices
    `no` and `ne` are set as the material parameters. With S!=1, a 
    :func:`uniaxial_order` is used to calculate actual material parameters.
    
    Parameters
    ----------
    director : ndarray
        A 4D array describing the director
    mask : ndarray, optional
        If provided, this mask must be a 3D bolean mask that define voxels where
        nematic is present. This mask is used to define the nematic part of the sample. 
        Volume not defined by the mask is treated as a host material. If mask is 
        not provided, all data points are treated as a director.
    no : float
        Ordinary refractive index
    ne : float
        Extraordinary refractive index 
    nhost : float
        Host refracitve index (if mask is provided)
    scale_factor : float
        The order parameter S obtained from the director length is scaled by this factor. 
        Optical anisotropy is then `epsa = S/scale_factor * (epse - epso)`.
    thickness : ndarray
        Thickness of layers (in pixels). If not provided, this defaults to ones.
        
    Returns
    -------
    optical_data : tuple
        A valid optical data tuple.
        
    """
    material = np.empty(shape = director.shape, dtype = FDTYPE)
    material[...] = refind2eps([no,no,ne])[None,...] 
    material = uniaxial_order(director2order(director)/scale_factor, material, out = material)
    
    if mask is not None:
        material[np.logical_not(mask),:] = refind2eps([nhost,nhost,nhost])[None,...] 
        
    if thickness is None:
        thickness = np.ones(shape = (material.shape[0],))
    return  thickness, material, director2angles(director)

def Q2data(tensor, mask = None, no = 1.5, ne = 1.6, nhost = None,scale_factor = 1., 
           biaxial = False, thickness = None):
    """Builds optical data from Q tensor data. 
    
    Parameters
    ----------
    tensor : (...,6) or (...,3,3) array
        Q tensor with elements Q[0,0], Q[1,1], Q[2,2], Q[0,1], Q[0,2], Q[1,2]
    mask : ndarray, optional
        If provided, this mask must be a 3D bolean mask that define voxels where
        nematic is present. This mask is used to define the nematic part of the sample. 
        Volume not defined by the mask is treated as a host material. If mask is 
        not provided, all data points are treated as a director.
    no : float
        Ordinary refractive index (when biaxial = False)
    ne : float
        Extraordinary refractive index (when biaxial = False)
    nhost : float
        Host refracitve index (if mask is provided)
    scale_factor : float
        The order parameter S obtained from the Q tensor is scaled by this factor. 
        Optical anisotropy is then `epsa = S/scale_factor *(epse - epso)`.
    biaxial : bool
        Describes whether data is treated as biaxial or converted to uniaxial (default).
        If biaxial, no describes the mean value of n1 and n2 refractive indices
        eigenavalues and ne = n3.
    thickness : ndarray, optional
        Thickness of layers (in pixels). If not provided, this defaults to ones.
    
    Returns
    -------
    optical_data : tuple
        A valid optical data tuple.
    """
    tensor = np.asarray(tensor)
    eps = Q2eps(tensor, no = no, ne = ne,scale_factor = scale_factor)
    
    shape = eps.shape[:-1]
    if eps.shape[-1] == 6:
        shape = shape + (3,) 
        
    material, epsa = eps2epsva(eps)
            
    if not bool(biaxial):
        material = uniaxial_order(1, material, out = material)
    
    if mask is not None:
        material[np.logical_not(mask),:] = refind2eps([nhost,nhost,nhost])[None,...] 
        
    if thickness is None:
        thickness = np.ones(shape = (material.shape[0],))
    return  thickness, material, epsa


def director2Q(director, order = 1.):
    """Computes Q tensor form the uniaxial director. 
    
    Parameters
    ----------
    director : (...,3) array
        Director vector. The length of the vector is the square of the order 
        parameter.
    order : float, optional
        In case director is normalized, this describes the order parameter.
        
    Returns
    -------
    Q : (...,6) array
        Q tensor with elements Q[0,0], Q[1,1], Q[2,2], Q[0,1], Q[0,2], Q[1,2]
    """
    S = director2order(director)
    n = director
    
    out = np.empty(S.shape + (6,), S.dtype)
    
    d = S/3*order

    out[...,0] = order*n[...,0]**2 - d
    out[...,1] = order*n[...,1]**2 - d
    out[...,2] = order*n[...,2]**2 - d
    out[...,3] = order*n[...,0]*n[...,1]
    out[...,4] = order*n[...,0]*n[...,2]
    out[...,5] = order*n[...,1]*n[...,2]
    
    return out

               
def Q2director(tensor, qlength = False):
    """Computes the director from tensor data 
    
    Parameters
    ----------
    tensor : (...,6) or (...,3,3) array
        Tensor data.
    qlength : bool
        Specifies whether the length of the director is set to the S parameter
        or not. If not, director is scalled to unity length.
        
    Returns
    -------
    Q : (...,6) array
        Q tensor with elements Q[0,0], Q[1,1], Q[2,2], Q[0,1], Q[0,2], Q[1,2]
    """
    qeig, r = tensor_eig(tensor) #sorted eigenvalues.. qeig[2] is for the main axis (director)
    S = qeig[...,2] * 3/2. #the S parameter of the uniaxial q tensor
    if qlength:
        return S[...,None] * r[...,2]
    else:
        return r[...,2]
            
def Q2eps(tensor, no = 1.5, ne = 1.6,scale_factor = 1., out = None):
    """Converts Q tensor to epsilon tensor
    
    Parameters
    ----------
    tensor : (...,6) or (...,3,3) ndarray
        A 4D array describing the Q tensor. If provided as a matrix, rhe elemnents are
        Q[0,0], Q[1,1], Q[2,2], Q[0,1], Q[0,2], Q[1,2]
    no : float
        Ordinary refractive index
    ne : float
        Extraordinary refractive index 
    scale_factor : float
        The order parameter S obtained from the Q tensor is scaled by this factor. 
        Optical anisotropy is then `epsa = S/scale_factor * (epse - epso)`.
    out : ndarray, optional
        Output array 
    
    Returns
    -------
    eps : ndarray
        Calculated epsilon tensor.
    """

    qtensor = np.asarray(tensor)

    if qtensor.shape[-1] == 6:
        qdiag = qtensor[...,0:3]
        qoff = qtensor[...,3:]
    elif qtensor.shape[-2:] == (3,3):
        mask_off = np.array(((False,True,True),(False,False,True),(False,False,False)))
        mask_diag = np.array(((True,False,False),(False,True,False),(False,False,True)))

        qdiag = qtensor[...,mask_diag]
        qoff = qtensor[...,mask_off]
    else:
        raise ValueError("input tensor must be an array of shape (...,6)")
    epse = ne**2
    epso = no**2
    if out is None:
        out = np.empty_like(qtensor)
    #: scaled anisotropy
    epsa = (epse-epso) / scale_factor
    #: mean epsilon
    epsm = (epso*2 + epse)/3.
    
    if qtensor.shape[-1] == 6:
        out[...,0:3] = epsa * qdiag + epsm
        out[...,3:] = epsa * qoff
    else:
        out[...,mask_diag] = epsa * qdiag + epsm
        out[...,mask_off] = epsa * qoff
        out[...,1,0] = out[...,0,1]
        out[...,2,0] = out[...,0,2]
        out[...,2,1] = out[...,1,2]
    return out
  
def eps2epsva(eps):
    """Computes epsilon eigenvalues (epsv) and rotation angles (epsa) from
    epsilon tensor of shape (...,6) or represented as a (...,3,3)
    
    Parameters
    ----------
    eps : (...,3,3) or (...,6) array
       Epsilon tensor. If provided as a (3,3) matrix, the elements are
       eps[0,0], eps[1,1], eps[2,2], eps[0,1], eps[0,2], eps[1,2]

    Returns
    -------
    epsv, epsa : ndarray, ndarray
        Eigenvalues and Euler angles arrays.
    """
    epsv, r = tensor_eig(eps)
    # r is in general complex for complex eps. But, if a complex tensor is a rotated diagonal,
    # the eigenvectors should be real. Test it here.
    atol = 1e-8 if CDTYPE == "complex128" else 1e-8
    rtol = 1e-5 if CDTYPE == "complex128" else 1e-5
    if not np.allclose(r,r.real, atol = atol, rtol = rtol):
        import warnings
        warnings.warn("Input tensor is not normal because eigevectors are not real. Results are unpredictive!", stacklevel = 2)
        
    return epsv, rotation_angles(r.real)

def epsva2eps(epsv,epsa):
    """Computes epsilon from eigenvalues (epsv) and rotation angles (epsa) 
    
    Parameters
    ----------
    epsv : (...,3) array
       Epsilon eigenvalues array.
    epsa : (...,3) array
       Euler angles array.

    Returns
    -------
    eps : ndarray
        Epsilon tensor arrays of shape (...,6).  The elements are
       eps[0,0], eps[1,1], eps[2,2], eps[0,1], eps[0,2], eps[1,2]
    """
    r = rotation_matrix(epsa)
    return rotate_diagonal_tensor(r,epsv)

        
def validate_optical_data(data, homogeneous = False):
    """Validates optical data.
    
    This function inspects validity of the optical data, and makes proper data
    conversions to match the optical data format. In case data is not valid and 
    it cannot be converted to a valid data it raises an exception (ValueError). 
    
    Parameters
    ----------
    data : tuple of optical data
        A valid optical data tuple.
    homogeneous : bool, optional
        Whether data is for a homogenous layer. (Inhomogeneous by defult)
    
    Returns
    -------
    data : tuple
        Validated optical data tuple. 
    """
    thickness, material, angles = data
    thickness = np.asarray(thickness, dtype = FDTYPE)
    if thickness.ndim == 0:
        thickness = thickness[None] #make it 1D
    elif thickness.ndim != 1:
        raise ValueError("Thickess dimension should be 1.")
    n = len(thickness)
    material = np.asarray(material)
    if np.issubdtype(material.dtype, np.complexfloating):
        material = np.asarray(material, dtype = CDTYPE)
    else:
        material = np.asarray(material, dtype = FDTYPE)
    if (material.ndim == 1 and homogeneous) or (material.ndim==3 and not homogeneous):
        material = np.broadcast_to(material, (n,)+material.shape)# np.asarray([material for i in range(n)], dtype = material.dtype)
    if len(material) != n:
        raise ValueError("Material length should match thickness length")
    if (material.ndim != 2 and homogeneous) or (material.ndim != 4 and not homogeneous):
        raise ValueError("Invalid dimensions of the material.")

    angles = np.asarray(angles, dtype = FDTYPE)
    if (angles.ndim == 1 and homogeneous) or (angles.ndim==3 and not homogeneous):
        angles = np.broadcast_to(angles, (n,)+angles.shape)
        #angles = np.asarray([angles for i in range(n)], dtype = angles.dtype)
    if len(angles) != n:
        raise ValueError("Angles length should match thickness length")
    if (angles.ndim != 2 and homogeneous) or (angles.ndim!=4 and not homogeneous):
        raise ValueError("Invalid dimensions of the angles.")

    if material.shape != angles.shape:
        raise ValueError("Incompatible shapes for angles and material")
 
    return thickness.copy(), material.copy(), angles.copy()
    
def raw2director(data, order = "zyxn", nvec = "xyz"):
    """Converts raw data to director array.
    
    Parameters
    ----------
    data : array
        Data array
    order : str, optional
        Data order. It can be any permutation of 'xyzn'. Defaults to 'zyxn'. It
        describes what are the meaning of axes in data.
    nvec : str, optional
        Order of the director data coordinates. Any permutation of 'x', 'y' and 
        'z', e.g. 'yxz', 'zxy'. Defaults to 'xyz'  
        
    Returns
    -------
    director : array
        A new array or same array (if no trasposing and data copying was made)
        
    Example
    -------
    
    >>> a = np.random.randn(10,11,12,3)
    >>> director = raw2director(a, "xyzn")
    """
    if order != "zyxn":
        #if not in zxyn order, then we must transpose data
        try:
            axes = (order.find(c) for c in "zyxn")
            axes = tuple((i for i in axes if i != -1))
            data = np.transpose(data, axes)
        except:
            raise ValueError("Invalid value for 'order'. Must be a permutation of 'xyzn' characters")
        
    if nvec != "xyz":
        index = {"x" : 0, "y": 1, "z" : 2}
        out = np.empty_like(data)
        for i,idn in enumerate(nvec):
            j = index.pop(idn)
            out[...,j] = data[...,i]
        return out
    else:
        return data    

def read_raw(file, shape, dtype, sep = "", endian = sys.byteorder):
    """Reads raw data from a binary or text file.
    
    Parameters
    ----------
    file : str or file
        Open file object or filename.
    shape : sequence of ints
        Shape of the data array, e.g., ``(50, 24, 34, 3)``
    dtype : data-type
        Data type of the raw data. It is used to determine the size of the items 
        in the file.
    sep : str
        Separator between items if file is a text file.
        Empty ("") separator means the file should be treated as binary.
        Spaces (" ") in the separator match zero or more whitespace characters.
        A separator consisting only of spaces must match at least one
        whitespace.
    endian : str, optional
        Endianess of the data in file, e.g. 'little' or 'big'. If endian is 
        specified and it is different than sys.endian, data is byteswapped. 
        By default no byteswapping is done.
    """  
    dtype = np.dtype(dtype)
    count = np.multiply.reduce(shape) * dtype.itemsize
    a = np.fromfile(file, dtype, count, sep)
    if endian == sys.byteorder:
        return a.reshape(shape)  
    elif endian not in ("little", "big"):
        raise ValueError("Endian should be either 'little' or 'big'")
    else:
        return a.reshape(shape).byteswap(True)
       
def _r3(shape):
    """Returns r vector array of given shape."""
    az, ay, ax = [np.arange(-l / 2. + .5, l / 2. + .5) for l in shape]
    zz,yy,xx = np.meshgrid(az,ay,ax, indexing = "ij")
    return xx, yy, zz

    
def sphere_mask(shape, radius, offset = (0,0,0)):
    """Returns a bool mask array that defines a sphere.
    
    The resulting bool array will have ones (True) insede the sphere
    and zeros (False) outside of the sphere that is centered in the compute
    box center.
    
    Parameters
    ----------
    shape : (int,int,int)
        A tuple of (nlayers, height, width) defining the bounding box of the sphere.
    radius: int
        Radius of the sphere in pixels.
    offset: (int, int, int), optional
        Offset of the sphere from the center of the bounding box. The coordinates
        are (x,y,z).
        
    Returns
    -------
    out : array
        Bool array defining the sphere.
    """
    xx, yy, zz = _r3(shape)
    r = ((xx-offset[0])**2 + (yy-offset[1])**2 + (zz--offset[2])**2) ** 0.5 
    mask = (r <= radius)
    return mask   

def nematic_droplet_director(shape, radius, profile = "r", retmask = False):
    """Returns nematic director data of a nematic droplet with a given radius.
    
    Parameters
    ----------
    shape : tuple
        (nz,nx,ny) shape of the output data box. First dimension is the 
        number of layers, second and third are the x and y dimensions of the box.
    radius : float
        radius of the droplet.
    profile : str, optional
        Director profile type. It can be a radial profile "r", or homeotropic 
        profile with director orientation specified with the parameter "x", "y",
        or "z", or as a director tuple e.g. (np.sin(0.2),0,np.cos(0.2)). Note that
        director length  defines order parameter (S=1 for this example).
    retmask : bool, optional
        Whether to output mask data as well
        
    Returns
    -------
    out : array or tuple of arrays 
        A director data array, or tuple of director mask and director data arrays.
    """
    
    nz, ny, nx = shape
    out = np.zeros(shape = (nz,ny,nx,3), dtype = FDTYPE)
    xx, yy, zz = _r3(shape)
    
    r = (xx**2 + yy**2 + zz**2) ** 0.5 
    mask = (r <= radius)
    m = np.logical_and(mask,r != 0.)
    rm = r[m]
    if profile == "r":
        out[...,0][m] = xx[m]/rm
        out[...,1][m] = yy[m]/rm
        out[...,2][m] = zz[m]/rm
    elif isinstance(profile, str):
        index = {"x": 0,"y": 1,"z": 2}
        try:
            i = index[profile]
            out[...,i][m] = 1.
        except KeyError:
            raise ValueError("Unsupported profile type!")
    else:
        try:
            x,y,z = profile
            out[...,0][m] = x
            out[...,1][m] = y
            out[...,2][m] = z
        except:
            raise ValueError("Unsupported profile type!")
            
    if retmask == True:
        return mask, out
    else: 
        return out
    
def cholesteric_director(shape, pitch, hand = "left"):
    """Returns a cholesteric director data.
    
    Parameters
    ----------
    shape : tuple
        (nz,nx,ny) shape of the output data box. First dimension is the 
        number of layers, second and third are the x and y dimensions of the box.
    pitch : float
        Cholesteric pitch in pixel units.
    hand : str, optional
        Handedness of the pitch; either 'left' (default) or 'right'

    Returns
    -------
    out : ndarray
        A director data array
    """
    nz, ny, nx = shape
    
    if hand == 'left':
        phi = -2*np.pi/pitch*np.arange(nz)
    elif hand == "right":
        phi = 2*np.pi/pitch*np.arange(nz)
    else:
        raise ValueError("Unknown handedness '{}'".format(hand))
    out = np.zeros(shape = (nz,ny,nx,3), dtype = FDTYPE)

    for i in range(nz):
        out[i,...,0] = np.cos(phi[i])
        out[i,...,1] = np.sin(phi[i])
    return out    

def nematic_droplet_data(shape, radius, profile = "r", no = 1.5, ne = 1.6, nhost = 1.5):
    """Returns nematic droplet optical_data.
    
    This function returns a thickness,  material_eps, angles, info tuple 
    of a nematic droplet, suitable for light propagation calculation tests.
    
    Parameters
    ----------
    shape : tuple
        (nz,nx,ny) shape of the stack. First dimension is the number of layers, 
        second and third are the x and y dimensions of the compute box.
    radius : float
        radius of the droplet.
    profile : str, optional
        Director profile type. It can be a radial profile "r", or homeotropic 
        profile with director orientation specified with the parameter "x", 
        "y", or "z".
    no : float, optional
        Ordinary refractive index of the material (1.5 by default)
    ne : float, optional
        Extraordinary refractive index (1.6 by default)
    nhost : float, optional
        Host material refractive index (1.5 by default)
        
    Returns
    -------
    out : tuple of length 3
        A (thickness, material_eps, angles) tuple of three arrays
    """
    mask, director = nematic_droplet_director(shape, radius, profile = profile, retmask = True)
    return director2data(director, mask = mask, no = no, ne = ne, nhost = nhost)


def cholesteric_droplet_data(shape, radius, pitch, hand = "left", no = 1.5, ne = 1.6, nhost = 1.5):
    """Returns cholesteric droplet optical_data.
    
    This function returns a thickness,  material_eps, angles, info tuple 
    of a cholesteric droplet, suitable for light propagation calculation tests.
    
    Parameters
    ----------
    shape : tuple
        (nz,nx,ny) shape of the stack. First dimension is the number of layers, 
        second and third are the x and y dimensions of the compute box.
    radius : float
        radius of the droplet.
    pitch : float
        Cholesteric pitch in pixel units.
    hand : str, optional
        Handedness of the pitch; either 'left' (default) or 'right'
    no : float, optional
        Ordinary refractive index of the material (1.5 by default)
    ne : float, optional
        Extraordinary refractive index (1.6 by default)
    nhost : float, optional
        Host material refractive index (1.5 by default)
        
    Returns
    -------
    out : tuple of length 3
        A (thickness, material_eps, angles) tuple of three arrays
    """
    director = cholesteric_director(shape, pitch, hand = hand)
    mask = sphere_mask(shape, radius)
    return director2data(director, mask = mask, no = no, ne = ne, nhost = nhost)    

@numba.guvectorize([(NF32DTYPE[:],NF32DTYPE[:]),(NF64DTYPE[:],NFDTYPE[:])], "(n)->()", cache = NUMBA_CACHE)
def director2order(data, out):
    """Converts director data to order parameter (square root of the length of the director)"""
    c = data.shape[0]
    if c != 3:
        raise TypeError("invalid shape")
    x = data[0]
    y = data[1]
    z = data[2]
    s = np.sqrt(x**2+y**2+z**2)
    out[0] = s**0.5

@numba.guvectorize([(NF32DTYPE[:],NF32DTYPE[:]),(NF64DTYPE[:],NFDTYPE[:])], "(n)->(n)", cache = NUMBA_CACHE)
def director2angles(data, out):
    """Converts director data to angles (yaw, theta phi)"""
    c = data.shape[0]
    if c != 3:
        raise TypeError("invalid shape")

    x = data[0]
    y = data[1]
    z = data[2]
    phi = np.arctan2(y,x)
    theta = np.arctan2(np.sqrt(x**2+y**2),z)
    #s = np.sqrt(x**2+y**2+z**2)
    out[0] = 0. #yaw = 0.
    out[1] = theta
    out[2] = phi

@numba.guvectorize([(NF32DTYPE[:],NF32DTYPE[:]),(NF64DTYPE[:],NFDTYPE[:])], "(n)->(n)", cache = NUMBA_CACHE)
def angles2director(data, out):
    """Converts angles data (yaw,theta,phi) to director (nx,ny,nz)"""
    c = data.shape[0]
    if c != 3:
        raise TypeError("invalid shape")

    s = 1.
    theta = data[1]
    phi = data[2]

    ct = np.cos(theta)
    st = np.sin(theta)
    cf = np.cos(phi)
    sf = np.sin(phi)
    out[0] = s*cf*st
    out[1] = s*sf*st
    out[2] = s*ct

def expand(data, shape, xoff = None, yoff = None, zoff = None, fill_value = 0.):
    """Creates a new scalar or vector field data with an expanded volume. 
    Missing data points are filled with fill_value. Output data shape
    must be larger than the original data.
    
    Parameters
    ----------
    data : array_like
       Input vector or scalar field data
    shape : array_like
       A scalar or length 3 vector that defines the volume of the output data
    xoff : int, optional
       Data offset value in the x direction. If provided, original data is 
       copied to new data starting at this offset value. If not provided, data 
       is copied symmetrically (default).
    yoff, int, optional
       Data offset value in the x direction. 
    zoff, int, optional
       Data offset value in the z direction.     
    fill_value: array_like
       A length 3 vector of default values for the border volume data points.
       
    Returns
    -------
    y : array_like
       Expanded ouput data
    """
    data = np.asarray(data)
    nz,nx,ny = shape
    if nz >= data.shape[0] and ny >= data.shape[1] and nx >= data.shape[2]:
        out = np.empty(shape = shape + data.shape[3:], dtype = data.dtype)
        out[...,:] = fill_value
        if xoff is None:
            xoff = (shape[1] - data.shape[1])//2
        if yoff is None:
            yoff = (shape[2] - data.shape[2])//2
        if zoff is None:
            zoff = (shape[0] - data.shape[0])//2
    
        out[zoff:data.shape[0]+zoff,yoff:data.shape[1]+yoff,xoff:data.shape[2]+xoff,...] = data
        return out 
    else:
        raise ValueError("Requested shape {} is not larger than original data's shape".format(shape))

_REFIND_DECL = [(NF32DTYPE[:],NF32DTYPE[:]), (NF64DTYPE[:],NFDTYPE[:]),(NC64DTYPE[:],NC64DTYPE[:]), (NC128DTYPE[:],NCDTYPE[:])]

@numba.njit(_REFIND_DECL, cache = NUMBA_CACHE)  
def _refind2eps(refind, out):
    out[0] = refind[0]**2
    out[1] = refind[1]**2
    out[2] = refind[2]**2

_REFIND_DECL = [NF32DTYPE(NF32DTYPE), NFDTYPE(NF64DTYPE),NC64DTYPE(NC64DTYPE), NCDTYPE(NC128DTYPE)]

@numba.vectorize(_REFIND_DECL, cache = NUMBA_CACHE)  
def refind2eps(refind):
    """Converts refractive index to epsilon"""
    return refind**2

_EPS_DECL = [(NF32DTYPE,NF32DTYPE[:],NF32DTYPE[:]), (NF64DTYPE, NF64DTYPE[:],NFDTYPE[:]),
             (NF32DTYPE,NC64DTYPE[:],NC64DTYPE[:]), (NF64DTYPE, NC128DTYPE[:],NCDTYPE[:])
             ]

@numba.njit(_EPS_DECL, cache = NUMBA_CACHE)
def _uniaxial_order(order, eps, out):
    if order >= 0.:
        m = (eps[0] + eps[1] + eps[2])/3.
        delta = eps[2] - (eps[0] + eps[1])/2.
        if order == 0.:
            eps1 = m
            eps3 = m
        else:
            eps1 = m - 1./3. *order * delta
            eps3 = m + 2./3. * order * delta
        eps2 = eps1
    else:
        eps1 = eps[0]
        eps2 = eps[1]
        eps3 = eps[2]

    out[0] = eps1
    out[1] = eps2
    out[2] = eps3

    
_EPS_DECL_VEC = [(NF32DTYPE[:],NF32DTYPE[:],NF32DTYPE[:]), (NF64DTYPE[:], NF64DTYPE[:],NFDTYPE[:]),
             (NF32DTYPE[:],NC64DTYPE[:],NC64DTYPE[:]), (NF64DTYPE[:], NC128DTYPE[:],NCDTYPE[:])
             ]
@numba.guvectorize(_EPS_DECL_VEC ,"(),(n)->(n)", cache = NUMBA_CACHE)
def uniaxial_order(order, eig, out):
    """
    uniaxial_order(order, eps)
    
    Calculates uniaxial (or isotropic) eigen tensor from a diagonal biaxial eigen tensor.
    
    Parameters
    ----------
    order : float
        The order parameter. 1.: uniaxial, 0.: isotropic. If order is negative
        no change is made (biaxial case). 
    eig : array
        Array of shape (...,3), the eigenvalues. The eigenvalue [2] is treated 
        as the extraordinary axis for the uniaxial order. 
    out :ndarray, optional
        Output array.
        
    Parameters
    ----------
    out : ndarray
        Effective eigenvalues based on the provided symmetry (order) argument    
    
    >>> np.allclose(uniaxial_order(0,[1,2,3.]) , (2,2,2)) 
    True
    >>> uniaxial_order(1,[1,2,3.], (1.5,1.5,3)) 
    True
    >>> uniaxial_order(-1,[1,2,3.], (1,2,3))  #negative, so no change
    True
    >>> np.allclose(uniaxial_order(0.5,[1,2,3.]), (1.75,1.75,2.5)) #scale accordingly
    True
    """
    assert eig.shape[0] in (3,)
    _uniaxial_order(order[0], eig, out)
       
def eig_symmetry(order, eig, out = None):
    """Takes the ordered diagonal values of the tensor and converts it to 
    uniaxial or isotropic tensor, or keeps it as biaxial.

    Broadcasting rules apply.
    
    Parameters
    ----------
    order : int or array 
        Integer describing the symmetry 0 : isotropic, 1 : uniaxial, 2 : biaxial.
        If specified as an array it mast be broadcastable. See :func:`uniaxial_order`.
    eig : array
        Array of shape (...,3), the eigenvalues. The eigenvalue [2] is treated 
        as the extraordinary axis for the uniaxial order. 
    out :ndarray, optional
        Output array.
        
    Parameters
    ----------
    out : ndarray
        Effective eigenvalues based on the provided symmetry (order) argument
        
    See Also
    --------
    :func:`.uniaxial_order` for scalled order adjustment.
    """
    order = np.asarray(order, int)
    mask = order > 1
    #if order is negative... it is not applied in uniaxial_order function... so it remains biaxial
    order[mask] = -1
    return uniaxial_order(order ,eig, out)  
    
MAGIC = b"dtms" #legth 4 magic number for file ID
VERSION = b"\x01"


#IOs fucntions
#-------------

def save_stack(file, optical_data):
    """Saves optical data to a binary file in ``.dtms`` format.
    
    Parameters
    ----------
    file : file, str
        File or filename to which the data is saved.  If file is a file-object,
        then the filename is unchanged.  If file is a string, a ``.dtms``
        extension will be appended to the file name if it does not already
        have one.
    optical_data: optical data tuple
        A valid optical data
    """    
    own_fid = False
    d,epsv,epsa = validate_optical_data(optical_data)
    try:
        if isinstance(file, str):
            if not file.endswith('.dtms'):
                file = file + '.dtms'
            f = open(file, "wb")
            own_fid = True
        else:
            f = file
        f.write(MAGIC)
        f.write(VERSION)
        np.save(f,d)
        np.save(f,epsv)
        if epsa is not None:
            np.save(f,epsa)
    finally:
        if own_fid == True:
            f.close()

def load_stack(file):
    """Load optical data from a file.
    
    Parameters
    ----------
    file : file, str
        The file to read.
    """
    own_fid = False
    try:
        if isinstance(file, str):
            f = open(file, "rb")
            own_fid = True
        else:
            f = file
        magic = f.read(len(MAGIC))
        if magic == MAGIC:
            if ord(f.read(1)) > ord(VERSION):
                raise OSError("This file was created with a more recent version of dtmm. Please upgrade your dtmm package!")
            d = np.load(f)
            epsv = np.load(f)
            if f.peek(1) == b"":
                #no more data to read.. epsa is not present
                epsa =  None
            else:
                epsa = np.load(f)
            return d, epsv, epsa
        else:
            raise OSError("Failed to interpret file {}".format(file))
    finally:
        if own_fid == True:
            f.close()

def filter_data(optical_data, wavelength, pixelsize, betamax = 1, symmetry = "isotropic"):
    d, epsv, epsa = optical_data
    k = k0(wavelength, pixelsize) 
    epsv, epsa = filter_epsva(epsv, epsa, k, betamax )
    if symmetry == "uniaxial":
        uniaxial_order(1., epsv, epsv)
    elif symmetry == "isotropic":
        uniaxial_order(0., epsv, epsv)
    elif symmetry != "biaxial":
        raise ValueError("Unknown symmetry.")    
    return d, epsv, epsa
       
def filter_eps(eps, k, betamax = 1):
    eps = np.moveaxis(eps,-1,-3)
    feps = fft2(eps)
    beta, phi = betaphi(feps.shape[-2:],k)
    mask = beta > betamax
    feps[...,mask] = 0.
    eps = ifft2(feps)
    eps = np.moveaxis(eps,-3,-1)
    return eps

def filter_epsva(epsv, epsa, k, betamax = 1):
    eps = epsva2eps(epsv,epsa)
    eps = filter_eps(eps, k, betamax)
    return eps2epsva(eps)

_symmetry_key_maps = {"isotropic" : 0, "uniaxial" : 1,  "biaxial" : 2}

def _symmetry_arg_to_int(arg):
    if isinstance(arg, str):
        try:
            return _symmetry_key_maps[arg]
        except KeyError:
            raise ValueError("Invalid symmetry argument.")
    elif arg in (0,1,2,-1):
        return arg
    else:
        raise ValueError("Invalid symmetry argument.")
        
def _parse_symmetry_argument(arg):
    if isinstance(arg, str):
        return _symmetry_arg_to_int(arg)
    else:
        try:
            return tuple((_parse_symmetry_argument(a) for a in arg))
        except TypeError:
            return _symmetry_arg_to_int(arg)
        

def effective_data(optical_data, symmetry = 0):
    """Builds effective data from the optical_data.
    
    The material epsilon is averaged over the layers.
    
    Parameters
    ----------
    optical_data : tuple
        A valid optical_data tuple.

    symmetry : str, int or array 
        Either 'isotropic' or 0,  'uniaxial' or 1 or 'biaxial' or 2 .
        Defines the symmetry of the effective layer. When set to 'isotropic', 
        the averaging is done so that the effective layer tensor is isotropic. 
        When set to 'uniaxial' the  effective layer tensor is an uniaxial medium. 
        If it is an array, it defines the symmetry of each individual layers
        independetly.
        
    Returns
    -------
    out : tuple
        A valid optical data tuple of the effective layers.
    """
    d, epsv,epsa = optical_data
    #Whic axes are used for averaging averaging
    axis = list(range(len(epsv.shape)-1))
    
    if symmetry in ("isotropic",0):
        #no need to convert from eigenframe to laboratory frame to do averaging, just average out the
        epsv = eig_symmetry(0, epsv).mean(tuple(axis))
        epsa = np.zeros_like(epsv)
    else:
        order = np.asarray(_parse_symmetry_argument(symmetry),int)
        
        if order.ndim != 0:
            if len(order) != len(d):
                raise ValueError("Shape of the symmetry argument is incompatible.")
            #do not average over thickness axis, so pop it out.
            axis.pop(0) 
            
        axis = tuple(axis) #must be a tuple for np.mean
        
        #convert to laboratory frame, and perform averaging
        eps0 = epsva2eps(epsv,epsa)
        eps = eps0.mean(axis)
        
        #convert back to eigenframe
        epsv, epsa = eps2epsva(eps)
        eig_symmetry(order, epsv, out = epsv)
     
    if epsv.ndim == 1:
        #must be same lengths as d, so repeat the matereial for each layer
        epsv = np.asarray((epsv,)*len(d)).copy()#make a copy, to have a contiguous layer
        epsa = np.asarray((epsa,)*len(d)).copy()
    
    return d, epsv, epsa
    
def tensor2matrix(tensor, out = None):
    """Converts  matrix to tensor
    
    Parameters
    ----------
    tensor : (...,6) array
       Input 3x3 array
    out : (...,3,3) array, optional
       Output array.
       
    Returns
    -------
    matrix : ndarray
        Matrix of shape (...,3,3).
    """
    tensor = np.asarray(tensor)
    if tensor.shape[-1:] != (6,):
        raise ValueError("Not a valid tensor.")
    if out is None:
        out = np.empty(shape = tensor.shape[:-1] + (3,3), dtype = tensor.dtype)
    for i in range(3):
        out[...,i,i] = tensor[i]
    out[...,0,1] = tensor[...,3]
    out[...,0,2] = tensor[...,4]
    out[...,1,2] = tensor[...,5]
    out[...,1,0] = tensor[...,3]
    out[...,2,0] = tensor[...,4]
    out[...,2,1] = tensor[...,5]   
    return out
    
def matrix2tensor(matrix, out = None):
    """Converts  matrix to tensor
    
    Parameters
    ----------
    matrix : (...,3,3) array
       Input 3x3 array
    matrix : (...,6) array, optional
       Output array.
       
    Returns
    -------
    tensor : ndarray
        Tensor of shape (...,6).
    """
    matrix = np.asarray(matrix)
    if matrix.shape[-2:] != (3,3):
        raise ValueError("Not a valid matrix.")
    if out is None:
        out = np.empty(shape = matrix.shape[:-2] + (6,), dtype = matrix.dtype)
    for i in range(3):
        out[...,i] = matrix[...,i,i]
    out[...,3] = matrix[...,0,1]
    out[...,4] = matrix[...,0,2]
    out[...,5] = matrix[...,1,2]
    return out
    
if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 19 10:32:48 2018

@author: andrej
"""

from __future__ import absolute_import, print_function, division

from dtmm.data import angles2director, validate_optical_data
from mpl_toolkits.mplot3d import Axes3D #needed in order to have projection 3d
import matplotlib.pyplot as plt
import numpy as np

def _r3(shape, center = False):
    """Returns r vector array of given shape."""
    nz,ny,nx = [l//2 for l in shape]
    if center == False:
        az, ay, ax = [np.arange(l) for l in shape]
    else:
        az, ay, ax = [np.arange(-l / 2. + .5, l / 2. + .5) for l in shape]
    zz, yy, xx = np.meshgrid(az, ax, ay, indexing = "ij")
    return xx, yy, zz

def _add_mask(mask,data, lim):
    if lim is not None:
        low, high = lim
        m = np.logical_and(data>= low, data<high)
        mask = np.logical_and(m, mask)
    return mask

def _mask(x,y,z, xlim = None, ylim = None, zlim = None):       
    mask = np.ones(x.shape, dtype = bool)
    mask = _add_mask(mask,x, xlim)
    mask = _add_mask(mask,y, ylim)
    mask = _add_mask(mask,z, zlim)
    return mask

#def plot_id(data, id = 0, center = False, xlim = None, 
#            ylim = None, zlim = None, fig = None, ax = None,):
#    """Plots material id of the optical data.
#    
#    Parameters
#    ----------
#    
#    deta : optical_data or material_id 
#        A valid optical data tuple, or material_id array
#    id : int or array_like of ints
#        material id or multiple material ids to plot. 
#    """
#    if isinstance(data, tuple):
#        d,material_id,eps, angles = data
#    else:
#        material_id = data
#    if material_id.ndim == 2:
#        material_id = material_id[None,...]
#    assert material_id.ndim == 3
#    index = np.asarray(id)
#    if index.ndim == 0:
#        index = index[None]
#    material_id = np.asarray(material_id,dtype = "uint32")
#    if fig is None:
#        fig = plt.figure()
#    if ax is  None:
#        ax = fig.add_subplot(111, projection='3d')
#        ax.set_aspect('equal')
#    xx, yy, zz = _r3(material_id.shape, center)
#    for i in index:
#        mask = (material_id == i)
#        mask = _add_mask(mask,xx, xlim)
#        mask = _add_mask(mask,yy, ylim)
#        mask = _add_mask(mask,zz, zlim)
#        xs = xx[mask]
#        ys = yy[mask]
#        zs = zz[mask]
#        ax.scatter(xs, ys, zs,depthshade = False, s = 1, marker = ".", label = i)
#        
#    ax.set_xlabel('X')
#    ax.set_ylabel('Y')
#    ax.set_zlabel('Z')
#    plt.legend()
#    return fig, ax


def unique_eps(material, nmax = 5):
    n = np.multiply.reduce(material.shape[0:-1])
    mat = material.reshape((n,material.shape[-1]))
    eps = mat[0]
    out = [eps]
    for i in range(nmax-1):
        mask = (mat[...,0] != eps[0]) | (mat[...,1] != eps[1]) | (mat[...,2] != eps[2])
        mat = mat[mask,:]
        try:
            eps = mat[0]
            out.append(eps)
        except IndexError:
            #no more uniqe eps values
            break
    return out
        

def plot_material(data, eps = None, center = False, xlim = None, 
            ylim = None, zlim = None, ax = None):
    """Plots material (in color) of the optical data.
    
    Parameters
    ----------
    
    data : optical_data or material (eps)
        A valid optical data tuple, eps array
    eps : array or None, optional
        Specifies which eps values to plot. If not given, all data is plotted.
    center : bool, optional
        Whether to view coordinates from the center of the box.
    xlim : (low, high) or None, optional
        A tuple describing the coordinate limits in the x direction (width).
    ylim : (low, high) or None, optional
        A tuple describing the coordinate limits in the y direction (height).
    zlim : (low, high) or None, optional
        A tuple describing the coordinate limits in the z direction (layer index).
    ax : matplotlib.axes or None, optional
        If specified, plot to ax.
    """
    if isinstance(data, tuple):
        d, material, angles = validate_optical_data(data)
    else:
        material = data
    if material.ndim == 3:
        material = material[None,...]
    material = np.asarray(material,dtype = "float32")
    
    assert material.ndim == 4
    
    if eps is None:
        eps = unique_eps(material)
    else:
        eps = np.asarray(eps)
        if eps.ndim == 1:
            eps = eps[None,:]
    
    if ax is  None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        try:
            #this does not work in recent versions of matplotlib
            ax.set_aspect('equal')
        except NotImplementedError:
            pass
    xx, yy, zz = _r3(material.shape[0:-1], center)
    
    #mmin, mmax  = material.min(), material.max()
    #if mmax != mmin:
    #    colors = (material-mmin)/(mmax-mmin)
    #else:
    #    colors = material-mmin
    
    for e in eps:
        mask = (material[...,0] == e[0]) & (material[...,1] == e[1]) & (material[...,2] == e[2])
        mask = _add_mask(mask,xx, xlim)
        mask = _add_mask(mask,yy, ylim)
        mask = _add_mask(mask,zz, zlim)
        xs = xx[mask]
        ys = yy[mask]
        zs = zz[mask]
        #cs = colors[mask,:]
        #ax.scatter(xs, ys, zs,depthshade = False, s = 1, marker = ".", c = cs)
        ax.scatter(xs, ys, zs,depthshade = False, s = 1, marker = ".", label = e)
        
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    plt.legend()

    return ax.figure, ax

def plot_director(director, fig = None, center = False, xlim = None, 
                  ylim = None, zlim = None, ax = None):
    """Plots director data.
    
    Parameters
    ----------
    
    director : optical_data or material (eps)
        A valid optical data tuple, eps array
    eps : array or None, optional
        Specifies which eps values to plot. If not given, all data is plotted.
    center : bool, optional
        Whether to view coordinates from the center of the box.
    xlim : (low, high) or None, optional
        A tuple describing the coordinate limits in the x direction (width).
    ylim : (low, high) or None, optional
        A tuple describing the coordinate limits in the y direction (height).
    zlim : (low, high) or None, optional
        A tuple describing the coordinate limits in the z direction (layer index).
    ax : matplotlib.axes or None, optional
        If specified, plot to ax.
    """
    director = np.asarray(director)
    if director.ndim == 3:
        director = director[None,...]
    assert director.ndim == 4
    
    if ax is  None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        try:
            ax.set_aspect('equal')
        except NotImplementedError:
            #for some reason nev versions of matplotlib wont allow you to set aspect
            pass
    # Make the grid
    xx, yy, zz = _r3(director.shape[0:3], center)
    mask = _mask(xx,yy,zz,xlim,ylim,zlim)
    
    # Direction data for the arrows
    u = director[...,0][mask]
    v = director[...,1][mask]
    w = director[...,2][mask]

    ax.quiver(xx[mask], yy[mask], zz[mask], u, v, w, length = 1., arrow_length_ratio = 0.3 ,pivot = "middle")#, length=0.1, normalize=True)
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')    
    return ax.figure, ax

def plot_angles(data, **kwargs):
    """Plots eps angles for optical data or angles data.
    
    Parameters
    ----------
    
    data : optical_data or material (eps)
        A valid optical data tuple, eps array
    eps : array or None, optional
        Specifies which eps values to plot. If not given, all data is plotted.
    center : bool, optional
        Whether to view coordinates from the center of the box.
    xlim : (low, high) or None, optional
        A tuple describing the coordinate limits in the x direction (width).
    ylim : (low, high) or None, optional
        A tuple describing the coordinate limits in the y direction (height).
    zlim : (low, high) or None, optional
        A tuple describing the coordinate limits in the z direction (layer index).
    ax : matplotlib.axes or None, optional
        If specified, plot to ax.
    """
    if isinstance(data, tuple):
        d,eps, angles = validate_optical_data(data)
    else:
        angles = data
    director = angles2director(angles)
    return plot_director(director, **kwargs)

__all__ = ["plot_material", "plot_angles", "plot_director"]
    
if __name__ == "__main__":
    import dtmm
    import dtmm
    data = dtmm.nematic_droplet_data((60, 128, 128), 
                    radius = 30, profile = "r", no = 1.5, ne = 1.7, nhost = 1.5)
    thickness, material_eps, angles = data

    fig = plot_material(material_eps, center = True, xlim = (-40,-20),ylim = (-10,10), zlim = (-10,10))
    fig = dtmm.plot_angles(angles, center = True, xlim = (-3,3), ylim = (-3,3), zlim = (-3,3))


"""
Jones calculus functions. You can use this module for simple normal incidence
jones calculus.

2x2 matrices
------------

* :func:`.jonesvec` : jones vector
* :func:`.mirror` : mirror matrix
* :func:`.polarizer` : polarizer matrix
* :func:`.retarder` : retarder matrix
* :func:`.quarter_waveplate` : quarter-wave plate matrix
* :func:`.half_waveplate` : half-wave plate  matrix
* :func:`.circular_polarizer` : circular polarizer matrix
* :func:`.linear_polarizer` : linear polarizer matrix

Conversion functions
--------------------

* :func:`.as4x4` : convert to 4x4 matrix for (...,4) field vector.
* :func:`.rotated_matrix` : to view the matrix in rotated frame.

Examples
--------

>>> j_in = jonesvec((1,1)) #input light polarized at 45
>>> r = retarder(0.2,0.4)
>>> p = linear_polarizer((1,0)) # x polarization
>>> c = circular_polarizer(+1) # left handed
>>> m = multi_dot([r,p,c])
>>> j_out = dotmv(m,j_in) #output light E field
>>> i = jones_intensity(j_out) #intensity
"""

from dtmm.conf import CDTYPE,FDTYPE
import numpy as np
from dtmm.rotation import rotation_matrix2
from dtmm.linalg import dotmv, dotmm, multi_dot

def jones_intensity(jones):
    """Computes intensity of the jones vector
    
    Parameters
    ----------
    jones : (...,2) array
        Jones vector.
    
    Returns
    -------
    intensity : (...) float array
        Computed intensity.
    """
    return ((jones * np.conj(jones)).sum(-1)).real

def jonesvec(pol, phi = 0., out = None):
    """Returns a normalized jones vector from an input length 2 vector., Additionaly,
    you can use this function to view the jones vector in rotated coordinate frame, 
    defined with a rotation angle phi. 
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    pol : (...,2) array
        Input jones vector. Does not need to be normalized. 
    phi : float or (...,1) array
        If jones vector must be view in the rotated frame, this parameter
        defines the rotation angle.
    out : ndarray
        Output array in which to put the result; if provided, it
        must have a shape that the inputs broadcast to.
        
    Returns
    -------
    vec : ndarray
        Normalized jones vector
    
    Example
    -------
    
    >>> jonesvec((1,1j)) #left-handed circuar polarization
    array([0.70710678+0.j        , 0.        +0.70710678j])
    
    In rotated frame
    
    >>> j1,j2 = jonesvec((1,1j), (np.pi/2,np.pi/4)) 
    >>> np.allclose(j1, (0.70710678j,-0.70710678))
    True
    >>> np.allclose(j2, (0.5+0.5j,-0.5+0.5j))
    True
    
    """
    pol = np.asarray(pol, CDTYPE)
    phi = np.asarray(phi)
    if pol.shape[-1] != 2:
        raise ValueError("Invalid input shape")
    norm = (pol[...,0] * pol[...,0].conj() + pol[...,1] * pol[...,1].conj())**0.5
    pol = pol/norm[...,np.newaxis]
    pol = np.asarray(pol, CDTYPE)
    r = rotation_matrix2(-phi)
    return dotmv(r, pol, out)

def mirror(out = None):
    """Returns jones mirror matrix.
    
    Parameters
    ----------
    out : ndarray, optional
        Output array in which to put the result; if provided, it
        must have a shape that the inputs broadcast to.
        
    Returns
    -------
    mat : ndarray
        Output jones matrix.      
    """
    m = np.array(((1,0),(0,-1)), dtype = CDTYPE)
    if out is not None:
        out[...,:,:] = m
        m = out
    return m

def retarder(phase, phi = 0., out = None):
    """Returns jones matrix of general phase retarder
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    phase : float or array
        Phase difference between the fast and slow axis of the retarder.
    phi : float or array 
        Fast axis orientation
    out : ndarray, optional
        Output array in which to put the result; if provided, it
        must have a shape that the inputs broadcast to.
        
    Returns
    -------
    mat : ndarray
        Output jones matrix.
    """
    phi = np.asarray(phi)
    phase = np.asarray(phase)
    
    c = np.cos(phi*2)
    s = np.sin(phi*2)
    
    cp = np.cos(phase/2.)
    sp = np.sin(phase/2.)
    
    m00 = (cp + 1j*sp*c) 
    m11 = (cp - 1j*sp*c) 
    
    m01 = 1j*sp * s 
    
    shape = m00.shape + (2,2)
    if out is None:
        out = np.empty(shape = shape, dtype = CDTYPE)
    else:
        assert out.shape == shape 
    
    out[...,0,0] = m00
    out[...,0,1] = m01
    out[...,1,0] = m01
    out[...,1,1] = m11
    return out     

def quarter_waveplate(phi = 0., out = None):
    """Returns jones quarter-wave plate matrix.
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    phi : float or array 
        Fast axis orientation
    out : ndarray, optional
        Output array in which to put the result; if provided, it
        must have a shape that the inputs broadcast to.
        
    Returns
    -------
    mat : ndarray
        Output jones matrix.        
    """
    return retarder(np.pi/2, phi,  out)  

def half_waveplate(phi = 0., out = None):
    """Returns jones half-wave plate matrix.
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    phi : float or array 
        Fast axis orientation
    out : ndarray, optional
        Output array in which to put the result; if provided, it
        must have a shape that the inputs broadcast to.
        
    Returns
    -------
    mat : ndarray
        Output jones matrix.    
    """
    return retarder(np.pi, phi, out)         

def polarizer(jones, out = None):
    """Returns jones polarizer matrix.
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    jones : (...,2) array
        Input normalized jones vector. Use :func:`.jonesvec` to generate jones vector
    out : ndarray, optional
        Output array in which to put the result; if provided, it
        must have a shape that the inputs broadcast to.
        
    Returns
    -------
    mat : ndarray
        Output jones matrix.
    
    Examples
    --------
    >>> pmat = polarizer(jonesvec((1,1))) #45 degree linear polarizer
    >>> np.allclose(pmat, linear_polarizer(np.pi/4)+0j )
    True
    >>> pmat = polarizer(jonesvec((1,-1))) #-45 degree linear polarizer
    >>> np.allclose(pmat, linear_polarizer(-np.pi/4)+0j )
    True
    >>> pmat = polarizer(jonesvec((1,1j))) #left handed circular polarizer
    >>> np.allclose(pmat, circular_polarizer(1) )
    True
    """
    jones = np.asarray(jones)
    assert jones.shape[-1] == 2
    shape = jones.shape + (2,)
    if out is None:
        out = np.empty(shape = shape, dtype = CDTYPE)
    else:
        assert out.shape == shape 
    c,s = jones[...,0], jones[...,1]
    
    out[...,0,0] = c*c.conj()
    out[...,0,1] = c*s.conj()
    out[...,1,0] = s*c.conj()
    out[...,1,1] = s*s.conj()
    return out    
    
def linear_polarizer(angle, out = None):
    """Return jones matrix for a polarizer.
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    angle : float or array
        Orientation of the polarizer.
    out : ndarray, optional
        Output array in which to put the result; if provided, it
        must have a shape that the inputs broadcast to.
        
    Returns
    -------
    mat : ndarray
        Output jones matrix.
    """
    angle = np.asarray(angle)
    shape = angle.shape + (2,2)
    if out is None:
        out = np.empty(shape = shape, dtype = FDTYPE)
    else:
        assert out.shape == shape 
    c = np.cos(angle)
    s = np.sin(angle)
    cs = c*s
    out[...,0,0] = c*c
    out[...,1,0] = cs
    out[...,0,1] = cs
    out[...,1,1] = s*s
    return out

polarizer_matrix = linear_polarizer

def circular_polarizer(hand, out = None):
    """Returns circular polarizer matrix.
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    hand : int or (...,1) array
        Handedness +1 (left-hand) or -1 (right-hand).
    out : ndarray, optional
        Output array in which to put the result; if provided, it
        must have a shape that the inputs broadcast to.
        
    Returns
    -------
    mat : ndarray
        Output jones matrix.
    """
    hand = np.asarray(hand)*0.5
    shape = hand.shape + (2,2)
    if out is None:
        out = np.empty(shape = shape, dtype = CDTYPE)
    else:
        assert out.shape == shape 
    out[...,0,0] = 0.5
    out[...,0,1] = -1j*hand
    out[...,1,0] = 1j*hand
    out[...,1,1] = 0.5   
    return out

def rotated_matrix(jmat, phi):
    """jones matrix viewed in the rotated frame, where phi is the rotation angle
    
    Performs (R.T).jmat.R where R is the rotation matrix.
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    jmat : (...,2,2) array
        Jones matrix
    phi : float
        Rotation angle.
        
    Returns
    -------
    out : (...,2,2) array
        Rotated jones matrix.
    
    """
    r = np.asarray(rotation_matrix2(phi),CDTYPE)
    jmat = np.asarray(jmat)
    rT = np.swapaxes(r,-1,-2)
    return multi_dot([rT,jmat,r])

def as4x4(jmat,  out = None):
    """Converts jones 2x2 matrix to eigenfield 4x4 matrix.
    
    Parameters
    ----------
    jmat : (...,2,2) array
        Jones matrix
    out : ndarray, optional
        Output array in which to put the result; if provided, it
        must have a shape that the inputs broadcast to.
        
    Returns
    -------
    mat : (...,4,4) ndarray
        Output jones matrix.
    """
    
    if out is None:
        shape = jmat.shape[:-2] + (4,4)
        out = np.zeros(shape, dtype = jmat.dtype)
    else:
        out[...] = 0.
    out[...,0::2,0::2] = jmat    
    #for back propagating waves, jones matrix is conjugate
    out[...,1::2,1::2] = np.conj(jmat)

    return out

if __name__ == "__main__":
    import doctest
    doctest.testmod()


"""
Field vector and field data polarization and phase retardation matrix functions and handling.

Polarization and retardation handling 
+++++++++++++++++++++++++++++++++++++

* :func:`.polarizer4x4` creates a polarizer matrix.
* :func:`.jonesmat4x4` creates a general jones 4x4 matrix from 2x2 jones patrix. 
* :func:`.mode_jonesmat4x4` creates a general jones 4x4 matrix for field data in fft space.  
* :func:`.ray_jonesmat4x4` creates a general jones 4x4 matrix for field data in real space.  

Convenience functions
+++++++++++++++++++++

* :func:`.apply_mode_jonesmat4x4` to apply the matrix to field data in fft space
* :func:`.apply_ray_jonesmat4x4` to apply the matrix to field data in real space
* :func:`.apply_jonesmat` to apply the matrix to field vector (1D).

"""
from __future__ import absolute_import, print_function, division

from dtmm.conf import cached_result, BETAMAX, FDTYPE, CDTYPE
from dtmm.wave import betaphi
from dtmm.tmm import alphaf,normalize_f, E2H_mat
from dtmm.linalg import  dotmf, inv, dotmm
from dtmm.fft import fft2, ifft2
#make this module a drop-in replacement for jones module
from dtmm.jones import polarizer, as4x4, rotated_matrix
from dtmm.jones import * 
import numpy as np

from dtmm.diffract import diffraction_alphaf

def as2x2(pmat, fmat, mode = +1, out = None):
    """Converts jones 4x4 matrix to 2x2 E-field matrix"""
    H = E2H_mat(fmat, mode = mode)
    out = dotmm(pmat[...,::2,1::2], H, out)
    out += pmat[...,::2,::2] 
    return out

def polarizer4x4(jones, fmat, out = None):
    """Returns a 4x4 polarizer matrix for applying in eigenframe.
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    jones : (...,2) array
        A length two array describing the jones vector in the eigenframe. 
        Jones vector should be normalized.
    fmat : (...,4,4) array
        A field matrix array of the isotropic medium.
    out : ndarray, optional
        Output array.

    Returns
    -------
    jmat : (...,4,4) array
        A 4x4 matrix for field vector manipulation in the eigenframe.   
        
    See Also
    --------
    ray_jonesmat4x4 : for applying a general matrix in the laboratory frame.
    
    Examples
    --------
    >>> f = dtmm.tmm.f_iso(n = 1.) 
    >>> jones = dtmm.jones.jonesvec((1,0)) 
    >>> pol_mat = polarizer4x4(jones, f) #x polarizer matrix

    """
    jmat = polarizer(jones)
    return jonesmat4x4(jmat, fmat, out)

def jonesmat4x4(jmat, fmat, out = None):
    """Returns a 4x4 jones matrix for applying in eigenframe.
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    jmat : (...,2,2) array
        A 2x2 jones matrix in the eigenframe. Any of matrices in :mod:`dtmm.jones` 
        can be used.
    fmat : (...,4,4) array
        A field matrix array of the isotropic medium.
    out : ndarray, optional
        Output array   
        
    Returns
    -------
    jmat : (...,4,4) array
        A 4x4 matrix for field vector manipulation in the eigenframe.  
        
    See Also
    --------
    ray_jonesmat4x4 : for applying the jones matrix in the laboratory frame.
    """
    fmat = normalize_f(fmat)
    fmati = inv(fmat)
    pmat = as4x4(jmat)    
    m = dotmm(fmat,dotmm(pmat,fmati, out = out), out = out)
    return m

@cached_result
def mode_jonesmat4x4(shape, k, jmat, epsv = (1.,1.,1.), 
                            epsa = (0.,0.,0.), betamax = BETAMAX):
    """Returns a mode polarizer for fft of the field data in the laboratory frame.
    
    This is the most general set of jones matrices for field data. It is meant
    to be used in FFT space.
    
    Parameters
    ----------
    shape : (int,int)
        Shape of the 2D crossection of the field data.
    k : float or array of floats
        Wavenumber at which to compute the mode matrices.
    jmat : (2,2) array
        A 2x2 jones matrix that needs to be converted to 4x4 mode matrices.
    epsv : array
        Medium epsilon eigenvalues
    epsa : array
        Medium epsilon euler angles
    betamax : float
        The betamax parameter
        
    Returns
    -------
    pmat : (:,:,4,4) array
        Output matrix. Shape of the matirx is shape + (4,4) or len(ks) + shape + (4,4) 
        if k is an array. 
        
    See Also
    --------
    ray_jonesmat4x4 : for applying the jones matrix in the real space.
    jonesmat4x4 : for applying the jones matrix in the eigenframe.
    """
    ks = np.asarray(k, FDTYPE)
    ks = abs(ks)
    epsv = np.asarray(epsv, CDTYPE)
    epsa = np.asarray(epsa, FDTYPE)
    beta, phi = betaphi(shape,ks)
    alpha, f = diffraction_alphaf(shape, ks, epsv = epsv, 
                            epsa = epsa, betamax = betamax)

    beta, phi = betaphi(shape,ks)
    
    #matrix viewed in the eigenframe
    jmat = rotated_matrix(jmat, phi)
    pmat = jonesmat4x4(jmat, f)
    return pmat

@cached_result
def mode_jonesmat2x2(shape, k, jmat, epsv = (1.,1.,1.), 
                            epsa = (0.,0.,0.), mode = +1, betamax = BETAMAX):
    """Returns a mode polarizer for fft of the field data in the laboratory frame.
    
    This is the most general set of jones matrices for E-field data. It is meant
    to be used in FFT space.
    
    Parameters
    ----------
    shape : (int,int)
        Shape of the 2D crossection of the field data.
    k : float or array of floats
        Wavenumber at which to compute the mode matrices.
    jmat : (2,2) array
        A 2x2 jones matrix that needs to be converted to 4x4 mode matrices.
    epsv : array
        Medium epsilon eigenvalues
    epsa : array
        Medium epsilon euler angles
    mode : int
        PRopagatin mode, either +1 or -1
    betamax : float
        The betamax parameter
        
    Returns
    -------
    pmat : (:,:,2,2) array
        Output matrix. Shape of the matirx is shape + (2,2) or len(ks) + shape + (2,2) 
        if k is an array. 
        
    See Also
    --------
    ray_jonesmat2x2 : for applying the jones matrix in the real space.
    """
    ks = np.asarray(k, FDTYPE)
    ks = abs(ks)
    epsv = np.asarray(epsv, CDTYPE)
    epsa = np.asarray(epsa, FDTYPE)
    beta, phi = betaphi(shape,ks)
    alpha, f = diffraction_alphaf(shape, ks, epsv = epsv, 
                            epsa = epsa, betamax = betamax)

    beta, phi = betaphi(shape,ks)
    
    #matrix viewed in the eigenframe
    jmat = rotated_matrix(jmat, phi)
    pmat = jonesmat4x4(jmat, f)
    return as2x2(pmat,f,mode)

@cached_result
def ray_jonesmat4x4(jmat, beta = 0, phi = 0, epsv = (1.,1.,1.), 
                            epsa = (0.,0.,0.),):
    """Returns a ray polarizer for field data in the laboratory frame.
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    jmat : (...,2,2) array
        A 2x2 jones matrix that needs to be converted to 4x4 mode matrices.
    beta : float
        The beta parameter of the beam.
    phi : float
        The phi parameter of the beam.
    epsv : (...,3) array
        Medium epsilon eigenvalues
    epsa : (...,3) array
        Medium epsilon euler angles
    betamax : float
        The betamax parameter
        
    Returns
    -------
    pmat : (...,4,4) array
        Output matrix.
        
    See Also
    --------
    mode_jonesmat4x4 : for applying the jones matrix in the fft space.
    jonesmat4x4 : for applying the jones matrix in the eigenframe.
    """
    epsv = np.asarray(epsv, CDTYPE)
    epsa = np.asarray(epsa, FDTYPE)
    beta = np.asarray(beta, FDTYPE)
    phi = np.asarray(phi, FDTYPE)
    
    alpha, f = alphaf(beta, phi, epsv, epsa)
    
    jmat = rotated_matrix(jmat, phi)
    pmat = jonesmat4x4(jmat, f)

    return pmat

@cached_result
def ray_jonesmat2x2(jmat, beta = 0, phi = 0, epsv = (1.,1.,1.), 
                            epsa = (0.,0.,0.),mode = +1):
    """Returns a ray polarizer for E-field data in the laboratory frame.
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    jmat : (...,2,2) array
        A 2x2 jones matrix that needs to be converted to 2x2 mode matrices.
    beta : float
        The beta parameter of the beam.
    phi : float
        The phi parameter of the beam.
    epsv : (...,3) array
        Medium epsilon eigenvalues
    epsa : (...,3) array
        Medium epsilon euler angles
    betamax : float
        The betamax parameter
        
    Returns
    -------
    pmat : (...,2,2) array
        Output matrix.
        
    See Also
    --------
    mode_jonesmat2x2 : for applying the jones matrix in the fft space.
    jonesmat2x2 : for applying the jones matrix in the eigenframe.
    """
    epsv = np.asarray(epsv, CDTYPE)
    epsa = np.asarray(epsa, FDTYPE)
    beta = np.asarray(beta, FDTYPE)
    phi = np.asarray(phi, FDTYPE)
    
    alpha, f = alphaf(beta, phi, epsv, epsa)
    
    jmat = rotated_matrix(jmat, phi)
    pmat = jonesmat4x4(jmat, f)
    
    return as2x2(pmat, f, mode)

def apply_mode_jonesmat(pmat,field, out = None):
    """Multiplies matrix with field data in fft space.
    
    Parameters
    ----------
    pmat : array 
        A 4x4 jones matrix of shape (...,:,:,4,4) for field data or
        2x2 jones matrix of shape (...,:,:,2,2) for E-field data or
    field :  array
        Field data array of shape (...,4,:,:) or (...,2,:,:).
    out : ndarray, optional
        If specified, the results are written here.
        
    Returns
    -------
    out : array
        Computed field array of shape (...,4,:,:) or (...,2,:,:).
    """
    fft = fft2(field, out = out)
    pfft = dotmf(pmat, fft ,out  = fft)
    return ifft2(fft, out = pfft)

def apply_ray_jonesmat(pmat,field, out = None):
    """Multiplies matrix with field data in real space.
    
    Parameters
    ----------
    pmat : array
        A 4x4 jones matrix of shape (...,4,4) for field data or
        2x2 jones matrix of shape (...,2,2) for E-field data or
    field : array
        Field data array of shape (...,4,:,:) or (...,2,:,:).
    out : array, optional
        If specified, the results are written here.
        
    Returns
    -------
    out : ndarray
        Computed field array of shape (...,4,:,:) or (...,2,:,:).
    """
    return dotmf(pmat, field ,out  = out)

def apply_jonesmat(pmat, field, out = None):
    """Multiplies a (2x2) or (4x4) jones matrix with field vector
    
    Parameters
    ----------
    pmat : ndarray
        A 4x4 jones matrix of shape (...,4,4) for field data or
        2x2 jones matrix of shape (...,2,2) for E-field data or
    field : array
        Field vector array of shape (...,4) or (...,2).
    out : array, optional
        If specified, the results are written here.
        
    Returns
    -------
    out : ndarray
        Computed field array of shape (...,4).
    """
    return dotmm(pmat, field ,out  = out)


import unittest
import numpy as np
import dtmm.data as data
import dtmm.rotation as rot
import dtmm.conf as conf

Rx = rot.rotation_matrix_x
Ry = rot.rotation_matrix_y
Rz = rot.rotation_matrix_z

class TestData(unittest.TestCase):
    
    def setUp(self):
        pass
    
    def test_rotate_director(self):
         """
         Tests the rotate_director function using rot90_director.
         This test only verifies 90 degree rotations since other rotations are not exact.
         Returns
         -------

         """
         # Strings representing the rotation
         rotation_strings = ("+x", "-2x", "+2y", "-y", "y", "z", "3z")
         # Rotation functions to use
         rotation_functions = (Rx, Rx, Ry, Ry, Ry, Rz, Rz)
         # Angles used to create rotation matrices
         angles = (np.pi/2, -np.pi, np.pi, -np.pi/2, np.pi/2, np.pi/2, -np.pi/2)

         # Number of vertical layers
         n_layers = 5
         # height of each layer
         heigh = 5
         # width of each layer
         width = 5

         # Create random test data of size
         test_data = np.random.randn(n_layers, heigh, width, 3)
         # Cast data to correct data type. Done in case single precision is being used
         test_data = np.asarray(test_data, dtype=conf.FDTYPE)

         # Iterate through each rotation test
         for rotation_str, rotation_function, angle in zip(rotation_strings, rotation_functions, angles):
             # Calculate exact result using rot90_director
             rotated_data_goal = data.rot90_director(test_data, rotation_str)
             # Create rotation matrix
             rotation_matrix = rotation_function(angle)
             # Rotate director, do not normalize length because initial data is not normalized
             rotated_data = data.rotate_director(rotation_matrix, test_data, norm = False)
             # Compare inner data, without boundaries 
             self.assertTrue(np.allclose(rotated_data_goal[1:-1,1:-1,1:-1], rotated_data[1:-1,1:-1,1:-1]))
        
if __name__ == "__main__":
    unittest.main()
